"""
https://en.wikipedia.org/wiki/Ribbon_diagram

https://behreajj.medium.com/scripting-curves-in-blender-with-python-c487097efd13


Carson, M.; Bugg, C. E. (1986),
"Algorithm for Ribbon Models of Proteins",
Journal of Molecular Graphics, 4 (2): 121-122,
doi:10.1016/0263-7855(86)80010-8.

"""

import bpy
from time import time
import numpy as np
from batoms.ribbon.protein import Protein
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)


def draw_curve_from_vertices_bezier(
    name,
    data,
    coll,
    extrude=None,
    bevel_depth=None,
    bevel_control=None,
    node_type="Principled BSDF",
    node_inputs=None,
    material_style="metallic",
    backface_culling=True,
):
    """ """
    from batoms.material import create_material

    vertices = data["vertices"]
    crv = bpy.data.curves.new(name, "CURVE")
    crv.dimensions = "3D"
    crv.resolution_u = 10
    crv.fill_mode = "FULL"
    crv.use_fill_caps = True
    spline = crv.splines.new(type="BEZIER")
    nvert = len(vertices)
    spline.bezier_points.add(nvert - 1)
    # vertices = np.append(vertices, np.zeros((nvert, 1)), axis = 1)
    vertices = vertices.reshape(-1, 1)
    spline.bezier_points.foreach_set("co", vertices)
    for i in range(nvert):
        spline.bezier_points[i].handle_right_type = "AUTO"
        spline.bezier_points[i].handle_left_type = "AUTO"
    if extrude:
        crv.extrude = extrude
    if bevel_depth:
        crv.bevel_depth = bevel_depth
    # Create bevel control curve.
    if bevel_control:
        crv.bevel_object = bevel_control
        crv.bevel_mode = "OBJECT"
    obj = bpy.data.objects.new(name, crv)
    #
    material = create_material(
        name,
        data["color"],
        node_type=node_type,
        node_inputs=node_inputs,
        material_style=material_style,
        backface_culling=backface_culling,
    )
    obj.data.materials.append(material)
    coll.objects.link(obj)


def draw_rope_from_vertices_nurbs(
    name,
    data,
    coll,
    extrude=None,
    bevel_depth=None,
    bevel_control=None,
    node_type="Principled BSDF",
    node_inputs=None,
    material_style="metallic",
    backface_culling=True,
):
    """ """
    from batoms.material import create_material

    vertices = data["vertices"]
    crv = bpy.data.curves.new(name, "CURVE")
    crv.dimensions = "3D"
    crv.resolution_u = 10
    crv.fill_mode = "FULL"
    crv.use_fill_caps = True
    crv.twist_mode = "Z_UP"  # 'TANGENT' #'Z_UP'
    crv.twist_smooth = 1
    spline = crv.splines.new(type="NURBS")
    # spline.use_endpoint_u = True
    nvert = len(vertices)
    spline.points.add(nvert - 1)
    vertices = np.append(vertices, np.ones((nvert, 1)), axis=1)
    vertices = vertices.reshape(-1, 1)
    # tilts = data['tilts']
    spline.points.foreach_set("co", vertices)
    # spline.points.foreach_set('tilt', tilts)
    # spline.order_u = 4
    # Set the main curve's bevel control to the bevel control curve.
    if extrude:
        crv.extrude = extrude
    if bevel_depth:
        crv.bevel_depth = bevel_depth
    # Create bevel control curve.
    if bevel_control:
        crv.bevel_object = bevel_control
        crv.bevel_mode = "OBJECT"
    obj = bpy.data.objects.new(name, crv)
    #
    material = create_material(
        name,
        data["color"],
        node_type=node_type,
        node_inputs=node_inputs,
        material_style=material_style,
        backface_culling=backface_culling,
    )
    obj.data.materials.append(material)
    coll.objects.link(obj)


def curve2mesh(name, vertices, resolution_u=20, coll=None):
    """
    Use Blender's curve to get b-spline
    resolution: number of vertices = (n(control points) - 1)*preview_U
    default preview_U  is 12
    """
    # tstart = time()
    crv = bpy.data.curves.new("%s-curve" % name, "CURVE")
    crv.dimensions = "3D"
    crv.resolution_u = resolution_u
    crv.fill_mode = "FULL"
    spline = crv.splines.new(type="NURBS")
    # spline.use_endpoint_u = True
    nvert = len(vertices)
    spline.points.add(nvert - 1)
    vertices = np.append(vertices, np.ones((nvert, 1)), axis=1)
    vertices = vertices.reshape(-1, 1)
    spline.points.foreach_set("co", vertices)
    obj = bpy.data.objects.new(name, crv)
    # if coll:
    # coll.objects.link(obj)
    me = obj.to_mesh()
    n = len(me.vertices)
    vertices = np.zeros(3 * n)
    # print('vertices: ', nvert, n)
    me.vertices.foreach_get("co", vertices)
    # print('curve2mesh: %s vertices, time: %s'%(nvert, (time() - tstart)))
    return vertices.reshape(-1, 3)


def draw_sheet_from_vertices_spline(
    name,
    data,
    coll,
    node_type="Principled BSDF",
    node_inputs=None,
    material_style="metallic",
    backface_culling=True,
    shade_smooth=True,
):
    """ """
    from batoms.material import create_material
    from batoms.ribbon.profile import build_mesh

    vertices = curve2mesh(
        "%s-vertices" % name, data["vertices"], data["resolution"], coll
    )
    sides = curve2mesh("%s-sides" % name, data["sides"], data["resolution"])
    normals = curve2mesh("%s-normals" % name, data["normals"], data["resolution"])
    profiles = data["profiles"]
    scales = data["scales"]
    vertices, faces = build_mesh(vertices, normals, sides, profiles, scales)
    me = bpy.data.meshes.new(name)
    me.from_pydata(vertices, [], faces)
    if shade_smooth:
        me.polygons.foreach_set("use_smooth", [True] * len(me.polygons))
    obj = bpy.data.objects.new(name, me)
    #
    material = create_material(
        name,
        data["color"],
        node_type=node_type,
        node_inputs=node_inputs,
        material_style=material_style,
        backface_culling=backface_culling,
    )
    obj.data.materials.append(material)
    coll.objects.link(obj)


class Ribbon:
    """ """

    def __init__(self, label, batoms=None, datas={}, update=False) -> None:
        self.label = label
        self.batoms = batoms
        self.protein = Protein(batoms)
        self.protein.setting(datas)
        if update:
            self.protein.update()

    @property
    def coll(self):
        return self.get_coll()

    def get_coll(self):
        return bpy.data.collections.get("%s_ribbon" % self.label)

    def draw_sheet(self):
        tstart = time()
        for name, sheet in self.protein.sheets.items():
            draw_sheet_from_vertices_spline(
                "sheet-%s" % name, sheet.as_dict(), self.coll, shade_smooth=False
            )
        logger.debug("draw sheet: %s" % (time() - tstart))
        self.batoms.selects["all"].show = False

    def draw_helix(self):
        tstart = time()
        for name, helix in self.protein.helixs.items():
            draw_sheet_from_vertices_spline(
                "helix-%s" % name, helix.as_dict(), self.coll, shade_smooth=True
            )
        logger.debug("draw helix: %s" % (time() - tstart))
        self.batoms.selects["all"].show = False

    def draw_turn(self):
        # Create bevel control curve.
        tstart = time()
        bpy.ops.curve.primitive_bezier_circle_add(radius=0.25, enter_editmode=False)
        bevel_control = bpy.context.active_object
        bevel_control.hide_set(True)
        bevel_control.hide_render = True
        bevel_control.data.name = bevel_control.name = "%s_turn_bevel" % self.label
        self.coll.objects.link(bevel_control)
        for name, turn in self.protein.turns.items():
            draw_rope_from_vertices_nurbs(
                "turn-%s" % name, turn.as_dict(), self.coll, bevel_control=bevel_control
            )
        logger.debug("draw turn: %s" % (time() - tstart))
        self.batoms.selects["all"].show = False

    def draw(self):
        tstart = time()
        self.draw_sheet()
        self.draw_helix()
        self.draw_turn()
        logger.debug("draw ribbon: %s" % (time() - tstart))



"""

1) set custom file folder.
2) use default startup of batoms.
3) use default preferences of batoms.
#TODO  batoms_setting_path_update
"""

import bpy
from bpy.types import AddonPreferences
from bpy.props import (
    BoolProperty,
    StringProperty,
    EnumProperty,
)
from batoms.install.pip_dependencies import has_module
from batoms.install import update
from batoms.logger import update_logging_level
import logging

logger = logging.getLogger(__name__)


# Enum property.
# Note: the getter/setter callback must use integer identifiers!
logging_level_items = [
    ("DEBUG", "DEBUG", "", 0),
    ("INFO", "INFO", "", 1),
    ("WARNING", "WARNING", "", 2),
    ("ERROR", "ERROR", "", 3),
    ("CRITICAL", "CRITICAL", "", 4),
]

dependencies = {
    "ase": "ase",
    "scikit-image": "skimage",
    "spglib": "spglib",
    "pymatgen": "pymatgen",
    "openbabel": "openbabel",
}


DEFAULT_GITHUB_ACCOUNT = "beautiful-atoms"
DEFAULT_REPO_NAME = "beautiful-atoms"
DEFAULT_PLUGIN_NAME = "batoms"


def get_plugin(key):
    """Helper function to get plugin
    Args:
        key (_type_): _description_
    """

    def getter(self):
        # self.get: returns the value of the custom property assigned to
        # key or default when not found
        value = self.bl_rna.properties[key].default
        return self.get(key, value)

    return getter


def set_plugin(key):
    """Helper function to set plugin

    Args:
        key (_type_): _description_
        value (_type_): _description_
    """

    def setter(self, value):
        import importlib

        self[key] = value
        plugin = importlib.import_module("batoms.plugins.{}".format(key))
        if value:
            plugin.register_class()
            logger.info("Enable {} plugin.".format(key))
        else:
            plugin.unregister_class()
            logger.info("Disnable {} plugin.".format(key))

    return setter


class BatomsDefaultPreference(bpy.types.Operator):
    """Update Batoms"""

    bl_idname = "batoms.use_batoms_preference"
    bl_label = "Use defatul preference of Batoms"
    bl_description = "Use startup file of Batoms"

    def execute(self, context):
        import pathlib
        import os

        batoms_asset_dir = os.path.join(
            pathlib.Path(__file__).parent.resolve(), "asset"
        )
        batoms_asset_dir = os.path.join(batoms_asset_dir, "libraries")

        bpy.context.scene.unit_settings.system = "NONE"
        #
        bpy.context.preferences.view.use_translate_new_dataname = False
        bpy.context.preferences.inputs.use_rotate_around_active = True
        bpy.context.preferences.inputs.use_zoom_to_mouse = True
        # For laptop
        bpy.context.preferences.inputs.use_emulate_numpad = True
        # For laptop without mouse
        bpy.context.preferences.inputs.use_mouse_emulate_3_button = True
        # bpy.context.window.workspace = bpy.data.workspaces['UV Editing']
        # theme
        bpy.context.preferences.themes[
            0
        ].view_3d.space.gradients.background_type = "LINEAR"
        bpy.context.preferences.themes[0].view_3d.space.gradients.high_gradient = (
            0.9,
            0.9,
            0.9,
        )
        bpy.context.preferences.themes[0].view_3d.space.gradients.gradient = (
            0.5,
            0.5,
            0.5,
        )
        bpy.ops.wm.save_userpref()
        # logger
        bpy.context.preferences.addons["batoms"].preferences.logging_level = "WARNING"
        # asset_libraries
        if "Batoms" not in bpy.context.preferences.filepaths.asset_libraries.keys():
            bpy.ops.preferences.asset_library_add(directory=batoms_asset_dir)
            bpy.context.preferences.filepaths.asset_libraries[-1].name = "Batoms"
        self.report({"INFO"}, "Set default preferences successfully!")
        return {"FINISHED"}


class BatomsDefaultStartup(bpy.types.Operator):
    """Update Batoms"""

    bl_idname = "batoms.use_batoms_startup"
    bl_label = "Use startup file of Batoms"
    bl_description = "Use defatul startup of Batoms"

    def execute(self, context):
        import os
        import pathlib

        addon_dir = pathlib.Path(__file__).parent.resolve()
        blend_dir = os.path.join(addon_dir, "data/startup.blend")
        bpy.ops.wm.open_mainfile(filepath=blend_dir, load_ui=True, use_scripts=True)
        ###################################################
        # Add additional settings to the startup file here
        ###################################################
        # viewport overlayrs
        for area in bpy.context.screen.areas:
            if area.type == "VIEW_3D":
                for space in area.spaces:
                    if space.type == "VIEW_3D":
                        space.overlay.show_extras = False
                        space.overlay.show_relationship_lines = False
                        break
        ###################################################
        bpy.ops.wm.save_homefile()
        self.report({"INFO"}, "Load default startup successfully!")
        # todo open preference again.
        # bpy.ops.screen.userpref_show('INVOKE_DEFAULT')
        # bpy.ops.preferences.addon_show(module="batoms")
        return {"FINISHED"}


class BatomsAddonPreferences(AddonPreferences):
    bl_idname = __package__

    def get_logging_level(self):
        items = self.bl_rna.properties["logging_level"].enum_items
        # self.get: returns the value of the custom property assigned to
        # key or default when not found
        return items[self.get("logging_level", 2)].value

    def set_logging_level(self, value):
        items = self.bl_rna.properties["logging_level"].enum_items
        item = items[value]
        level = item.identifier
        # we need to update both the preference and the logger
        self["logging_level"] = level
        # Set the logging level for all child loggers of "batoms"
        update_logging_level()
        # Note the following logging info might not emit
        # if global level is higher than INFO
        logger.info("Set logging level to: {}".format(level))

    def batoms_setting_path_update(self, context):
        import os

        if os.name == "posix":  # Linux
            cmds = ["export", "BATOMS_SETTING_PATH={}".format(self.batoms_setting_path)]
        if os.name == "nt":  # Windows
            cmds = ["setx", "BATOMS_SETTING_PATH {}".format(self.batoms_setting_path)]
        logger.debug(update.subprocess_run(cmds))

    ase: BoolProperty(
        name="ASE installed",
        description="ASE package installed",
        default=False,
    )

    skimage: BoolProperty(
        name="scikit-image installed",
        description="scikit-image package installed",
        default=False,
    )

    spglib: BoolProperty(
        name="spglib installed",
        description="spglib package installed",
        default=False,
    )

    pymatgen: BoolProperty(
        name="pymatgen installed",
        description="pymatgen package installed",
        default=False,
    )

    openbabel: BoolProperty(
        name="Openbabel installed",
        description="openbabel package installed",
        default=False,
    )

    batoms_setting_path: StringProperty(
        name="Custom Setting Path",
        description="Custom Setting Path",
        default="",
        subtype="FILE_PATH",
        update=batoms_setting_path_update,
    )

    logging_level: EnumProperty(
        name="Logging Level",
        items=logging_level_items,
        get=get_logging_level,
        set=set_logging_level,
        default=2,
    )

    isosurface: BoolProperty(
        name="isosurface",
        description="Enable isosurface plugin",
        get=get_plugin("isosurface"),
        set=set_plugin("isosurface"),
        default=True,
    )

    molecular_surface: BoolProperty(
        name="molecular_surface",
        description="Enable molecular_surface plugin",
        get=get_plugin("molecular_surface"),
        set=set_plugin("molecular_surface"),
        default=True,
    )

    real_interaction: BoolProperty(
        name="real_interaction",
        description="Enable real_interaction plugin",
        get=get_plugin("real_interaction"),
        set=set_plugin("real_interaction"),
        default=False,
    )

    magres: BoolProperty(
        name="magres",
        description="Enable magres plugin",
        get=get_plugin("magres"),
        set=set_plugin("magres"),
        default=True,
    )

    highlight: BoolProperty(
        name="highlight",
        description="Enable highlight plugin",
        get=get_plugin("highlight"),
        set=set_plugin("highlight"),
        default=True,
    )

    cavity: BoolProperty(
        name="cavity",
        description="Enable cavity plugin",
        get=get_plugin("cavity"),
        set=set_plugin("cavity"),
        default=True,
    )

    crystal_shape: BoolProperty(
        name="crystal_shape",
        description="Enable crystal_shape plugin",
        get=get_plugin("crystal_shape"),
        set=set_plugin("crystal_shape"),
        default=True,
    )

    lattice_plane: BoolProperty(
        name="lattice_plane",
        description="Enable lattice_plane plugin",
        get=get_plugin("lattice_plane"),
        set=set_plugin("lattice_plane"),
        default=True,
    )

    template: BoolProperty(
        name="template",
        description="Enable template plugin",
        get=get_plugin("template"),
        set=set_plugin("template"),
        default=True,
    )

    def draw(self, context):
        layout = self.layout

        layout.label(text="Welcome to Batoms!")
        # Check Blender version
        if bpy.app.version_string < "3.0.0":
            box = layout.box().column()
            box.label(text="Warning: Batoms need Blender version > 3.0.0.")

        box = layout.box().column()
        row = box.row(align=True)
        row.operator("batoms.update", icon="FILE_REFRESH")
        layout.label(text="Use default setting.")
        box = layout.box().column()
        row = box.row(align=True)
        row.operator(
            "batoms.use_batoms_startup", text="Use startup", icon="FILE_REFRESH"
        )
        row.operator(
            "batoms.use_batoms_preference", text="Use Preferences", icon="FILE_REFRESH"
        )
        layout.separator()

        box = layout.box().column()
        box.label(text="Dependencies:")
        #
        for package, modname in dependencies.items():
            if not has_module(modname):
                op = box.operator(
                    "batoms.pip_install_package",
                    icon="IMPORT",
                    text="Install {}".format(package),
                )
                op.package = package
                op.modname = modname
            else:
                setattr(self, modname, True)
                box.prop(self, modname, text=package)
        #
        layout.separator()
        split = layout.split()
        col = split.column()
        col.label(text="Custom Plugins")
        col.prop(self, "highlight")
        col.prop(self, "isosurface")
        col.prop(self, "molecular_surface")
        col.prop(self, "crystal_shape")
        col.prop(self, "lattice_plane")
        col.prop(self, "cavity")
        col.prop(self, "magres")
        col.prop(self, "real_interaction")
        col.prop(self, "template")
        col = split.column()
        # custom folder
        layout.separator()
        box = layout.box().column()
        box.label(text="Custom Settings")
        box.prop(self, "logging_level")
        box.prop(self, "batoms_setting_path")


classes = [
    BatomsDefaultPreference,
    BatomsDefaultStartup,
    BatomsAddonPreferences,
    update.BatomsUpdateButton,
]


def register_class():
    from bpy.utils import register_class

    for cls in classes:
        register_class(cls)


def unregister_class():
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(cls)



import bpy


def test_position(au):
    from time import time

    tstart = time()
    au.segments = [6, 6]
    au.repeat([10, 10, 10])
    au.repeat([5, 5, 5])
    t = time() - tstart
    assert t < 100
    print("Repeat time: {:1.2f}".format(t))
    # get position
    tstart = time()
    positions = au.positions
    t = time() - tstart
    assert t < 4
    print("Get positions time: {:1.2f}".format(t))
    # set position
    tstart = time()
    au["Au"].positions = positions
    t = time() - tstart
    assert t < 4
    print("Set positions time: {:1.2f}".format(t))


def test_scatter_and_gather_attribute():
    from ase.build import bulk
    import numpy as np
    from batoms import Batoms
    from time import time

    bpy.ops.batoms.delete()
    au = bulk("Au")
    au *= [5, 5, 5]
    au *= [10, 10, 10]
    # single value
    d0 = np.zeros((len(au)))
    d2 = np.zeros((len(au), 2))
    au.set_array("d0d", d0)
    au.set_array("d1d", d2)
    au = Batoms("au", from_ase=au)
    tstart = time()
    au.get_attribute("d0d")
    t = time() - tstart
    print("Gatther data for data (1,): {:1.2f}".format(t))
    assert t < 1
    tstart = time()
    au.get_attribute("d1d")
    t = time() - tstart
    print("Gatther data for data (2,): {:1.2f}".format(t))
    assert t < 2



from .bond import Bond

from . import (
    bpy_data,
    ops,
    gui,
    gui_slicebonds,
)

__all__ = ["Bond", "register_class", "unregister_class"]

classes_bpy_data = [
    # internal data first
    bpy_data.BondSetting,
    bpy_data.Bond,
    gui.BondProperties,
    gui_slicebonds.BondProperties,
]

classes = [
    ops.BondPairAdd,
    ops.BondPairRemove,
    ops.BondModify,
    ops.BondDraw,
    ops.BondOrderAutoSet,
    ops.BondShowHydrogenBond,
    ops.BondShowSearch,
    gui.VIEW3D_PT_Batoms_bond,
    gui.BATOMS_MT_bond_pair_context_menu,
    gui.BATOMS_UL_bond_pair,
    gui.BATOMS_PT_bond_pair,
    gui_slicebonds.Bond_PT_prepare,
]


def register_class():
    from bpy.types import Collection, Object, Scene
    from bpy.props import PointerProperty
    from bpy.utils import register_class

    for cls in classes_bpy_data:
        register_class(cls)
    # attach to blender internal data
    Collection.Bbond = PointerProperty(name="Bbond", type=bpy_data.Bond)
    Object.Bbond = PointerProperty(name="Bbond", type=bpy_data.BondSetting)
    Scene.Bbond = PointerProperty(type=gui_slicebonds.BondProperties)
    for cls in classes:
        register_class(cls)


def unregister_class():
    from bpy.types import Collection, Object, Scene
    from bpy.utils import unregister_class

    del Collection.Bbond
    del Object.Bbond
    del Scene.Bbond
    for cls in reversed(classes):
        unregister_class(cls)
    for cls in reversed(classes_bpy_data):
        unregister_class(cls)



import bpy
from bpy.props import (
    StringProperty,
    BoolProperty,
    IntProperty,
    FloatProperty,
    EnumProperty,
    FloatVectorProperty,
    CollectionProperty,
)

from batoms.internal_data import Base


class MagresSetting(Base):
    """ """

    flag: BoolProperty(name="flag", default=False)
    name: StringProperty(name="name")
    label: StringProperty(name="label", default="batoms")
    type: EnumProperty(
        name="type",
        description="Surface type",
        items=(
            ("MS", "MS", ""),
            ("CS", "CS", ""),
        ),
        default="MS",
    )
    scale: FloatProperty(name="scale", soft_min=0.01, soft_max=1, default=0.01)
    resolution: FloatProperty(name="resolution", soft_min=0.2, soft_max=2, default=0.5)
    select: StringProperty(name="select", default="all")
    color: FloatVectorProperty(
        name="color", size=4, subtype="COLOR", min=0, max=1, default=[0, 1, 1, 0.5]
    )

    def as_dict(self) -> dict:
        setdict = {
            "flag": self.flag,
            "label": self.label,
            "name": self.name,
            "color": self.color[:],
            "scale": self.scale,
            "resolution": self.resolution,
            "select": self.select,
            "material_style": self.material_style,
        }
        return setdict

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Name   select     scale   resolution    color  \n"
        s += "{:6s}  {:6s}  {:1.3f}  {:1.3f}  [{:1.2f}  {:1.2f}  {:1.2f}   {:1.2f}] \n".format(
            self.name,
            self.select,
            self.scale,
            self.resolution,
            self.color[0],
            self.color[1],
            self.color[2],
            self.color[3],
        )
        s += "-" * 60 + "\n"
        return s


class Magres(bpy.types.PropertyGroup):
    """This module defines the Magres properties to extend
    Blender’s internal data.

    """

    active: BoolProperty(name="active", default=False)
    settings: CollectionProperty(name="MagresSetting", type=MagresSetting)

    ui_list_index: IntProperty(name="ui_list_index", default=0)

    def as_dict(self) -> dict:
        setdict = {}
        return setdict



import bpy
from bpy.types import Operator
from bpy_extras.io_utils import ImportHelper, ExportHelper
from bpy.props import (
    BoolProperty,
    StringProperty,
    EnumProperty,
)
from batoms.batoms import Batoms
from batoms.bio.bio import read


class IMPORT_OT_batoms(Operator, ImportHelper):
    bl_idname = "batoms.import"
    bl_label = "Import"
    bl_options = {"REGISTER", "UNDO"}

    filename_ext = ".xyz"

    label: StringProperty(name="label", default="", description="Name of batoms.")

    render: BoolProperty(
        name="Add a default render", default=True, description="Do you need a render?"
    )

    model_style: EnumProperty(
        name="Type",
        description="Choose model",
        items=(
            ("0", "Space-filling", "Use ball"),
            ("1", "Ball-and-stick", "Use ball and stick"),
            ("2", "Polyhedral", "Use polyhedral"),
            ("3", "Stick", "Use stick"),
        ),
        default="0",
    )

    def draw(self, context):
        layout = self.layout
        layout.label(text="Import a Structure")
        box = layout.box()
        col = box.column()
        col.prop(self, "label")
        col.prop(self, "render")
        #
        box = layout.box()
        row = box.row()
        row.label(text="Structure model")
        row = box.row()
        col = row.column()
        col.prop(self, "model_style")

    def execute(self, context):
        inputfile = bpy.path.abspath(self.filepath)
        if self.label == "":
            batoms = read(inputfile)
        else:
            batoms = read(inputfile, label=self.label)
        batoms.model_style = self.model_style
        return {"FINISHED"}


class EXPORT_OT_batoms(Operator, ExportHelper):
    bl_idname = "batoms.export"
    bl_label = "Export"
    bl_options = {"REGISTER", "UNDO"}

    filename_ext = ".xyz"

    def draw(self, context):
        layout = self.layout
        layout.label(text="Export a Structure")

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj:
            return obj.batoms.type != "OTHER"
        else:
            return False

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=obj.batoms.label)
        batoms.write(self.filepath)
        context.view_layer.objects.active = obj
        return {"FINISHED"}


def menu_func_import_batoms(self, context):
    lay = self.layout
    lay.operator(IMPORT_OT_batoms.bl_idname, text="Batoms file (xyz, cif, pdb, ...)")


def menu_func_export_batoms(self, context):
    lay = self.layout
    lay.operator(EXPORT_OT_batoms.bl_idname, text="Batoms file (xyz, cif, pdb, ...)")



"""
"""
import bpy
from batoms.base.collection import Setting, tuple2string
from batoms.data import default_polyhedras

import logging

logger = logging.getLogger(__name__)


class PolyhedraSettings(Setting):
    """
    PolyhedraSetting object

    The PolyhedraSetting object store the polyhedra information.

    Parameters:

    label: str
        The label define the batoms object that a Setting belong to.

    """

    def __init__(
        self,
        label,
        batoms=None,
        polyhedras=None,
        polyhedrasetting=None,
    ) -> None:
        Setting.__init__(self, label, coll_name="%s" % label)
        self.name = "Bpolyhedra"
        self.batoms = batoms
        self.polyhedras = polyhedras
        if len(self) == 0:
            self.add_species_list(self.batoms.species.keys())
        if polyhedrasetting is not None:
            for key, data in polyhedrasetting.items():
                self[key] = data

    def build_materials(self, sp, node_inputs=None):
        """ """
        from batoms.material import create_material

        name = "%s_%s" % (self.label, sp["species"])
        if name in bpy.data.materials:
            mat = bpy.data.materials.get(name)
            bpy.data.materials.remove(mat, do_unlink=True)
        mat = create_material(
            name,
            color=sp["color"],
            node_inputs=node_inputs,
            material_style=sp["material_style"],
            backface_culling=False,
        )
        # logger.debug('build polyhedra materials: {}'.format(sp['color']))
        return mat

    def __setitem__(self, index, setdict):
        """
        Set properties
        # TODO can not set color by setting['C'].color = [1, 0, 0, 1]
        """
        name = tuple2string(index)
        subset = self.find(name)
        if subset is None:
            subset = self.bpy_setting.add()
        subset.species = index
        subset.name = name
        for key, value in setdict.items():
            setattr(subset, key, value)
        subset.label = self.label
        subset.flag = True
        self.build_materials(self[name].as_dict())

    def set_default(self, species_props):
        """ """
        for sp, props in species_props.items():
            if sp not in default_polyhedras:
                continue
            props["color"] = (
                props["color"][0],
                props["color"][1],
                props["color"][2],
                0.8,
            )
            self[sp] = props

    def add(self, key, value=None):
        props = self.batoms.species.species_props[key]
        if value:
            props.update(value)
        props["color"] = (props["color"][0], props["color"][1], props["color"][2], 0.8)
        props["species"] = key
        self[key] = props

    def add_species_list(self, species_list, only_default=True):
        for sp in species_list:
            self.add_species(sp, only_default)

    def add_species(self, name, only_default=True):
        species_props = self.batoms.species.species_props[name]
        if only_default and species_props["element"] not in default_polyhedras:
            return
        self.add(name, species_props)

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Center                color         width \n"
        for p in self.bpy_setting:
            s += "{:10s}   ".format(p.species)
            s += "[{:1.2f}  {:1.2f}  {:1.2f}  {:1.2f}]   ".format(
                p.color[0], p.color[1], p.color[2], p.color[3]
            )
            s += "{:1.3f} \n".format(p.width)
        s += "-" * 60 + "\n"
        return s

    @property
    def materials(self):
        return self.get_materials()

    def get_materials(self):
        materials = {}
        index = 0
        for sp in self:
            name = "%s_%s" % (self.label, sp.species)
            mat = bpy.data.materials.get(name)
            materials[sp.species] = [mat, index]
            index += 1
        return materials



import bpy
from bpy.types import Operator
from bpy.props import (
    StringProperty,
    IntProperty,
)
import bmesh
from batoms import Batoms
from ase import Atoms
from ase.build.rotate import rotation_matrix_from_points
from ase.build import rotate
from ase.data import covalent_radii, chemical_symbols
from ase.geometry import get_distances
import numpy as np
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)


molecules = {
    "C": Atoms(
        symbols=["C", "H", "H", "H", "H"],
        positions=[
            [0, 0, 0],
            [0, 0, -1.09],
            [0, 1.024, 0.36],
            [-0.89, -0.51, 0.36],
            [0.89, -0.51, 0.36],
        ],
    ),
    "N": Atoms(
        symbols=["N", "H", "H", "H"],
        positions=[[0, 0, 0], [0, 0, 1.02], [0.975, 0, 0.289], [-0.385, 0.896, 0.289]],
    ),
    "O": Atoms(
        symbols=["O", "H", "H"], positions=[[0, 0, 0], [0, 0, -0.969], [0.94, 0, 0.234]]
    ),
    "S": Atoms(
        symbols=["S", "H", "H"],
        positions=[[0, 0, 0], [0, 0.001, -1.364], [1.364, 0.003, 0]],
    ),
    "F": Atoms(symbols=["F"], positions=[[0, 0, 0]]),
    "Cl": Atoms(symbols=["Cl"], positions=[[0, 0, 0]]),
    "Br": Atoms(symbols=["Br"], positions=[[0, 0, 0]]),
}


def edit_atom(batoms, indices, element):
    """Replace by element

    Args:
        element (str): The target element
    """
    logger.debug("replace atoms")
    positions = batoms.positions
    species = batoms.arrays["species"]
    bondlists = batoms.bonds.bondlists
    removeH = []
    for i in indices:
        # copy mol
        mol = molecules[element].copy()
        # number of H atoms to be added
        nh = len(mol) - 1
        # how many atoms (nbond) are connected to atoms i
        bai0 = np.where((bondlists[:, 0] == i))[0]
        bai1 = np.where((bondlists[:, 1] == i))[0]
        bai = np.append(bondlists[bai0, 1], bondlists[bai1, 0]).astype(int)
        nbond = len(bai)
        sps = species[bai]
        indh = bai[np.where(sps == "H")[0]]
        # difference between nbond and nh
        dbond = nbond - nh
        if dbond >= 0:
            # nbond >= nh,
            # 1. no more new H will be add
            # 2. dh old H will be remove
            dh = min(len(indh), dbond)
            removeH.extend(indh[:dh].tolist())
            mol = mol[0:1]
            mol[0].position = positions[i]
        else:
            # nbond < nh,
            # 1. dh new H will be add
            # 2. no more old H will be remove
            dh = min(nh, -dbond)
            if nbond == 1:
                # second nearest neighbour need be search
                j = bai[0]
                a1 = mol[0].position - mol[1].position
                a2 = positions[i] - positions[j]
                bondlength = (
                    covalent_radii[chemical_symbols.index(element)]
                    + covalent_radii[chemical_symbols.index(species[j])]  # noqa: W503
                )
                mol.positions += positions[j] + a2 / np.linalg.norm(a2) * bondlength
                # how many atoms (nbond) are connected to atoms i
                baj0 = np.where((bondlists[:, 0] == j))[0]
                baj1 = np.where((bondlists[:, 1] == j))[0]
                baj = np.append(bondlists[baj0, 1], bondlists[baj1, 0]).astype(int)
                baj = baj[np.where(baj != i)[0]]
                b2 = np.cross([0, 0, 1], a2)
                rotate(mol, a1, a2, [1, 0, 0], b2, center=mol[0].position)
                # find another verctor
                maxd = 0
                maxi = 0
                for i in range(20):
                    angle = i * 9
                    mol1 = mol.copy()
                    mol1.rotate(angle, a2, center=mol[0].position)
                    p1 = mol1.positions[2:]
                    p2 = positions[baj]
                    D, D_len = get_distances(p1, p2)
                    maxd1 = np.min(D_len)
                    if maxd1 > maxd:
                        maxd = maxd1
                        maxi = i
                mol.rotate(maxi * 9, a2, center=mol[0].position)
                del mol[[1]]
                # find minimized overlap
            else:
                baj = np.insert(bai, 0, i, axis=0)
                # minimize_rotation_and_translation from ASE
                p = mol.positions[: (nbond + 1), :]
                p0 = positions[baj]
                # centeroids to origin
                c = np.mean(p, axis=0)
                p -= c
                c0 = np.mean(p0, axis=0)
                p0 -= c0
                # Compute rotation matrix
                R = rotation_matrix_from_points(p.T, p0.T)
                mol.set_positions(np.dot(mol.positions, R.T) + c0)
                del mol[range(1, (nbond + 1))]
        # rotate mol
        batoms.add_atoms(
            {"species": mol.get_chemical_symbols(), "positions": mol.positions}
        )
    indices.extend(removeH)
    batoms.delete(indices)
    batoms.model_style = 1


class MolecueEditElement(Operator):
    bl_idname = "batoms.molecule_edit_atom"
    bl_label = "Edit Atoms"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Edit Atoms"

    element: StringProperty(name="Element", default="C", description="element")
    bond_order: IntProperty(name="Bond order", default=1, description="bond order")

    @classmethod
    def poll(cls, context):
        return context.mode in {"EDIT_MESH"} and context.object.batoms.type != "OTHER"

    def execute(self, context):
        obj = context.object
        data = obj.data
        if data.total_vert_sel > 0:
            bm = bmesh.from_edit_mesh(data)
            indices = [
                s.index for s in bm.select_history if isinstance(s, bmesh.types.BMVert)
            ]
            self.report({"INFO"}, "%s atoms were replaced" % len(indices))
            batoms = Batoms(label=obj.batoms.label)
            edit_atom(batoms, indices, self.element)
            bpy.context.view_layer.objects.active = batoms.obj
            bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}



import bpy


def test_cell(ch4):
    """Cell panel"""
    import numpy as np

    bpy.context.view_layer.objects.active = ch4.obj
    # model_style
    assert np.isclose(bpy.context.scene.batoms.cell.cell_a0, 0)
    bpy.context.scene.batoms.cell.cell_a0 = 3
    assert np.isclose(bpy.context.scene.batoms.cell.cell_a0, 3)
    assert np.isclose(ch4.cell[0, 0], 3)



import bpy
import os
import numpy as np
from batoms.base.collection import BaseCollection
from batoms.render.light import Lights
from batoms.render.camera import Camera
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)


subcollections = ["light"]


class Render(BaseCollection):
    def __init__(
        self,
        label="batoms",
        batoms=None,
        viewport=[0, 0, 1],
        engine="EEVEE",
        output="batoms.png",
        animation=False,
        gpu=False,
        run_render=True,
        resolution=[1000, 758],
        transparent=True,
        #  compute_device_type='CUDA',
        studiolight="Default",
        samples=64,
    ):
        """Rendering Batoms object using blender.

        Object to render batoms object. A Render object has
        one camera, a list of lights.

        Parameters:

        label: str
            Name for the collection in Blender.
        batoms: None or Batoms object
        viewport: array
            The direction of the viewport windows.
        engine: str
            enum in ['BLENDER_WORKBENCH', 'BLENDER_EEVEE', 'CYCLES']
        output: str:
            filepath for the output image
        animation: bool
        gpu: bool
        run_render: bool
        transparent: bool
        resolution: list of 2 ints
        compute_device_type: str
            enum in ['NONE', 'CUDA', 'OPENGL', 'METAL', 'OPTIX'], depending
            on computer.
        studiolight: str
            enum in ['Default', 'basic.sl', 'outdoor.sl',
                    'paint.sl', 'rim.sl', 'studio.sl']
        samples: int
            Default 64
        """
        self.label = label
        self.coll_name = "%s_render" % label
        BaseCollection.__init__(self, coll_name=self.coll_name)
        self.batoms = batoms
        self.camera_name = "%s_camera" % self.label
        # self.compute_device_type = compute_device_type
        coll = bpy.data.collections.get(self.coll_name)
        # load from collection
        self.output = output
        if coll and coll.Brender.flag:
            self.lights = Lights(label)
            self.camera = Camera(label)
            # coll.Brender.viewport = viewport
            # coll.Brender.animation = animation
            # coll.Brender.run_render = run_render
        else:
            self.set_collection(viewport, animation=animation, run_render=run_render)
            self.transparent = transparent
            self.resolution = resolution
            self.gpu = gpu
            self.engine = engine
            self.studiolight = studiolight
            self.samples = samples
            self.camera = Camera(label, coll=self.coll)
            self.lights = Lights(label)
            self.lights.add("Default", lock_to_camera=True)

    def set_collection(self, viewport, animation=False, run_render=True):
        """
        build main collection and its child collections.
        """
        if bpy.data.collections.get(self.coll_name):
            raise Exception("Failed, the name %s already in use!" % self.coll_name)
        coll = bpy.data.collections.new(self.coll_name)
        self.scene.collection.children.link(coll)
        for sub_name in subcollections:
            subcoll = bpy.data.collections.new("%s_%s" % (self.label, sub_name))
            coll.children.link(subcoll)
        coll.Brender.flag = True
        coll.Brender.viewport = viewport
        coll.Brender.animation = animation
        coll.Brender.run_render = run_render

    def set(self, **kwargs):
        """
        Set parameters like set(key1=value1, key2=value2, ...).
        """
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def engine(self):
        return self.get_engine()

    @engine.setter
    def engine(self, engine):
        self.set_engine(engine)

    def get_engine(self):
        return self.scene.render.engine

    def set_engine(self, engine):
        if engine.upper() == "EEVEE":
            engine = "BLENDER_EEVEE"
        elif engine.upper() == "WORKBENCH":
            engine = "BLENDER_WORKBENCH"
        elif engine.upper() == "CYCLES":
            self.scene.cycles.use_denoising = True
        self.scene.render.engine = engine.upper()

    @property
    def compute_device_type(self):
        return self.get_compute_device_type()

    @compute_device_type.setter
    def compute_device_type(self, compute_device_type):
        self.set_compute_device_type(compute_device_type)

    def get_compute_device_type(self):
        return bpy.context.preferences.addons["cycles"].preferences.compute_device_type

    def set_compute_device_type(self, compute_device_type):
        bpy.context.preferences.addons[
            "cycles"
        ].preferences.compute_device_type = compute_device_type.upper()

    @property
    def viewport(self):
        return np.array(self.coll.Brender.viewport)

    @viewport.setter
    def viewport(self, viewport):
        if viewport is not None:
            # viewport = viewport/np.linalg.norm(viewport)
            self.coll.Brender.viewport = viewport
            self.update_camera()
            self.update_light()

    @property
    def distance(self):
        return self.coll.Brender.distance

    @distance.setter
    def distance(self, distance):
        if distance is not None:
            self.coll.Brender.distance = distance
            self.update_camera()
            self.update_light()

    @property
    def center(self):
        return np.array(self.coll.Brender.center)

    @center.setter
    def center(self, center):
        if center is not None:
            self.coll.Brender.center = center
            self.update_camera()
            self.update_light()

    @property
    def gpu(self):
        return self.get_gpu()

    @gpu.setter
    def gpu(self, gpu):
        self.set_gpu(gpu)

    def get_gpu(self):
        return self.coll.Brender.gpu

    def set_gpu(self, gpu):
        self.coll.Brender.gpu = gpu
        if gpu:
            self.scene.cycles.device = "GPU"
            # bpy.context.preferences.addons["cycles"].preferences.get_devices()
            cprefs = bpy.context.preferences.addons["cycles"].preferences
            for device in cprefs.devices:
                device["use"] = 1  # Using all devices, include GPU and CPU
            # auto set compute_device_type, depend on the computer
            for compute_device_type in ("METAL", "CUDA", "OPENCL", "NONE"):
                try:
                    cprefs.compute_device_type = compute_device_type
                    break
                except TypeError:
                    pass

    @property
    def run_render(self):
        return self.coll.Brender.run_render

    @run_render.setter
    def run_render(self, run_render):
        self.coll.Brender.run_render = run_render

    @property
    def animation(self):
        return self.coll.Brender.animation

    @animation.setter
    def animation(self, animation):
        self.coll.Brender.animation = animation

    @property
    def studiolight(self):
        return bpy.context.scene.display.shading.studio_light

    @studiolight.setter
    def studiolight(self, studiolight):
        bpy.context.scene.display.shading.studio_light = "%s" % studiolight

    @property
    def frame(self):
        return bpy.context.scene.display.shading.studio_light

    @frame.setter
    def frame(self, frame):
        self.scene.frame_set(frame)

    @property
    def samples(self):
        return self.scene.cycles.samples

    @samples.setter
    def samples(self, samples):
        self.scene.cycles.samples = samples

    @property
    def transparent(self):
        return self.scene.render.film_transparent

    @transparent.setter
    def transparent(self, transparent):
        self.scene.render.film_transparent = transparent

    @property
    def resolution(self):
        return [self.scene.render.resolution_x, self.scene.render.resolution_y]

    @resolution.setter
    def resolution(self, resolution):
        self.scene.render.resolution_x = int(resolution[0])
        self.scene.render.resolution_y = int(resolution[1])

    @property
    def use_motion_blur(self):
        if self.engine == "BLENDER_EEVEE":
            return bpy.context.scene.eevee.use_motion_blur
        elif self.engine == "BLENDER_CYCLES":
            return bpy.context.scene.cycles.use_motion_blur

    @use_motion_blur.setter
    def use_motion_blur(self, use_motion_blur):
        """
        'use_motion_blur': False,
        'motion_blur_position': 'START',
        'motion_blur_steps': 10,
        'motion_blur_shutter': 30.0,
        """
        bpy.context.scene.eevee.use_motion_blur = use_motion_blur
        bpy.context.scene.eevee.motion_blur_position = "START"
        bpy.context.scene.eevee.motion_blur_steps = 10
        bpy.context.scene.eevee.motion_blur_shutter = 30.0
        bpy.context.scene.cycles.use_motion_blur = use_motion_blur
        bpy.context.scene.cycles.motion_blur_position = "START"
        bpy.context.scene.cycles.motion_blur_shutter = 30.0

    def set_viewport_distance_center(self, center=None, padding=None, canvas=None):
        """
        Calculate canvas and direction
        """
        batoms = self.batoms
        if padding is None:
            padding = max(batoms.size) + 0.5
        if center is None:
            center = batoms.get_center_of_geometry()
        self.center = center
        if canvas is None:
            width, height, depth = batoms.get_canvas_box(
                direction=self.viewport, padding=padding
            )
        else:
            width = canvas[0]
            height = canvas[1]
            depth = canvas[2]
        if self.distance < 0:
            self.distance = max(10, depth)
        self.update_camera(width, height, depth / 2)
        self.update_light()

    def update_camera(self, width=None, height=None, depth=0):
        if width is not None:
            self.camera.ortho_scale = max(width, height)
            # sensor_with/lens = width/distance
            self.camera.lens = (
                self.camera.obj.data.sensor_width * (self.distance - depth) / width
            )
            self.resolution = [self.resolution[0], self.resolution[0] * height / width]
        self.camera.location = self.center + self.viewport * self.distance
        self.camera.look_at = self.center

    def update_light(self):
        """
        Calculate canvas and direction
        """
        from batoms.utils.butils import lock_to

        # plane
        # light
        for name, light in self.lights.lights.items():
            if light.lock_to_camera:
                lock_to(self.lights["Default"].obj, self.camera.obj)
            else:
                lock_to(light.obj, None)
                light.direction = light.direction
                # calculate direction in origial axis

    def init(self, batoms=None):
        if batoms is not None:
            self.batoms = batoms
        bpy.context.scene.camera = self.camera.obj
        self.distance = -1
        self.set_viewport_distance_center(center=None, padding=None, canvas=None)

    def run(self, batoms, center=None, padding=None, canvas=None, **kwargs):
        """
        render the model and export result
        """
        self.batoms = batoms
        self.set(**kwargs)
        self.set_viewport_distance_center(center=center, canvas=canvas, padding=padding)
        bpy.context.scene.camera = self.camera.obj
        #
        directory = os.path.split(self.output)[0]
        if directory and not os.path.exists(directory):
            os.makedirs(directory)
        self.scene.render.image_settings.file_format = "PNG"
        if self.run_render:
            self.scene.frame_end = self.batoms.nframe
            self.scene.render.filepath = "{0}".format(self.output)
            bpy.ops.render.render(write_still=1, animation=self.coll.Brender.animation)
        else:
            print("saving to {0}.blend".format(self.output))
            bpy.ops.wm.save_as_mainfile(
                "EXEC_SCREEN", filepath="{0}.blend".format(self.output)
            )

    def export(self, filename="blender-ase.obj"):
        if filename.split(".")[-1] == "obj":
            bpy.ops.export_scene.obj(filename)
        if filename.split(".")[-1] == "x3d":
            bpy.ops.export_scene.x3d(filename)
        if filename.split(".")[-1] == "xyz":
            self.export_xyz(filename)

    def render_move_camera(self, filename, loc1, loc2, n):
        from batoms.utils import getEquidistantPoints

        locs = getEquidistantPoints(loc1, loc2, n)
        i = 0
        for loc in locs:
            bpy.data.objects["Camera"].location = loc
            self.render(self, filename + "_{0:3d}".format(i))
            i += 1

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s += "Label:   %s \n" % (self.label)
        if self.batoms is not None:
            s += "Batoms:   %s \n" % (self.batoms.label)
        s += "Engine:   %s \n" % (self.engine)
        s += "Viewport: %s \n" % (self.viewport)
        s += "Camera:   type = %s, location = %s, ortho_scale = %s \n" % (
            self.camera.type,
            self.camera.location,
            self.camera.ortho_scale,
        )
        s += "Light: \n"
        s += self.lights.__repr__()
        s += "-" * 60 + "\n"
        return s



from .ribbon import Ribbon

from . import (
    bpy_data,
)

__all__ = ["Ribbon"]

classes = [
    # internal data first
    bpy_data.SheetSetting,
    bpy_data.HelixSetting,
    bpy_data.TurnSetting,
    bpy_data.Protein,
    # ops.ProteinAdd,
    # ops.ProteinRemove,
    # ops.ProteinDraw,
    # ui_list.BATOMS_MT_sheet_context_menu,
    # ui_list.BATOMS_UL_sheet,
    # ui_list.BATOMS_PT_sheet,
    # ui_list.BATOMS_MT_helix_context_menu,
    # ui_list.BATOMS_UL_helix,
    # ui_list.BATOMS_PT_helix,
    # ui_list.BATOMS_MT_turn_context_menu,
    # ui_list.BATOMS_UL_turn,
    # ui_list.BATOMS_PT_turn,
]


def register_class():
    from bpy.types import Collection, Object
    from bpy.props import PointerProperty
    from bpy.utils import register_class

    for cls in classes:
        register_class(cls)
    # attach to blender internal data
    Collection.Bprotein = PointerProperty(name="Bprotein", type=bpy_data.Protein)
    Object.Bprotein = PointerProperty(name="Bprotein", type=bpy_data.Protein)


def unregister_class():
    from bpy.types import Collection, Object
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(cls)

    del Collection.Bprotein
    del Object.Bprotein



import bpy
from ase.io import read
from batoms import Batoms


def test_animation_molecule():
    bpy.ops.batoms.delete()
    deca = read("../tests/datas/deca_ala_md.xyz", index=":")
    deca = Batoms("deca", from_ase=deca, load_trajectory=True)
    deca.model_style = 1
    deca.nframe > 1


def test_animation_crystal():
    bpy.ops.batoms.delete()
    atoms = read("../tests/datas/tio2_10.xyz", index=":")
    tio2 = Batoms("tio2", from_ase=atoms, load_trajectory=True)
    tio2.model_style = 2
    tio2.boundary = 0.01
    tio2.nframe == 10



import bpy
import numpy as np


def test_render(ch4):
    """Render panel"""
    ch4.cell = [1, 2, 3]
    bpy.context.view_layer.objects.active = ch4.obj
    # bpy.ops.batoms.render_add()
    ch4.render.init()
    assert ch4._render is not None
    # viewport
    ch4.render.viewport = [1, 0, 0]
    assert np.isclose(bpy.context.scene.batoms.render.viewport_x, 1)
    bpy.context.scene.batoms.render.viewport_y = 2
    assert ch4.render.viewport[1] == 2
    # camera
    ch4.render.camera.type = "ORTHO"
    assert bpy.context.scene.batoms.render.camera_type == "ORTHO"
    # Some GUI tests can only be tested manually by opening Blender
    # bpy.context.scene.batoms.render.camera_type = "PERSP"
    # assert ch4.render.camera.type == 'PERSP'
    ch4.render.camera.ortho_scale = 20
    assert np.isclose(bpy.context.scene.batoms.render.ortho_scale, 20)
    # bpy.context.scene.batoms.render.ortho_scale = 5
    # assert np.isclose(ch4.render.camera.ortho_scale, 5)
    # light
    ch4.render.lights["Default"].direction = [1, 0, 0]
    assert np.isclose(bpy.context.scene.batoms.render.light_direction_x, 1)
    ch4.render.lights["Default"].energy = 20
    assert np.isclose(bpy.context.scene.batoms.render.energy, 20)
    # bpy.context.scene.batoms.render.energy = 5
    # assert np.isclose(ch4.render.lights["Default"].energy, 5)


def test_batom(ch4):
    """Batom panel"""
    bpy.context.view_layer.objects.active = ch4.obj
    # scale
    assert np.isclose(bpy.context.scene.batoms.batoms.scale, ch4[0].scale)
    bpy.context.scene.batoms.batom.scale = 1
    assert np.isclose(ch4[0].scale, bpy.context.scene.batoms.batoms.scale)



import bpy


def test_batoms_measure(ch4):
    """Create a molecule use GUI ASE"""
    bpy.ops.object.mode_set(mode="OBJECT")
    ch4.obj.data.vertices.foreach_set("select", [1, 1, 0, 0, 0])
    bpy.ops.object.mode_set(mode="EDIT")
    bpy.ops.batoms.measure()



import bpy
import pytest
from ase.build import bulk
from batoms import Batoms
from ase.build import graphene_nanoribbon


@pytest.mark.skip(reason="Need to run manualy by mouse")
def test_force_field():
    bpy.ops.batoms.delete()
    gnr = graphene_nanoribbon(2, 2, type="armchair", saturated=True, vacuum=3.5)
    gnr.pbc = False
    gnr = Batoms("gnr", from_ase=gnr)


@pytest.mark.skip(reason="Need to run manualy by mouse")
def test_force_field_al():
    bpy.ops.batoms.delete()
    al = Batoms("al", from_ase=bulk("Al"))
    al = al * [1, 20, 1]
    al.pbc = False



"""
Adds an object mode tool to the toolbar.
# TODO Not working in  background mode,
fialed to register the keymaps.
"""

from bpy.types import WorkSpaceTool


class BatomsCell(WorkSpaceTool):
    bl_space_type = "VIEW_3D"
    bl_context_mode = "OBJECT"  # "OBJECT" or "EDIT_MESH"

    bl_idname = "batoms.apply_cell"
    bl_label = "Cell"
    bl_description = "Apply new cell parameters."
    # bl_icon = "ops.generic.select_circle"
    bl_icon = "ops.transform.resize"
    bl_widget = None

    bl_keymap = (
        (
            "batoms.apply_cell",
            {"type": "RIGHTMOUSE", "value": "PRESS"},
            {"properties": []},
        ),
        # ("batoms.apply_cell", {"type": "LEFTMOUSE", "value": "PRESS", "ctrl": True},
        #  {"properties": []}),
    )

    def draw_settings(context, layout, tool):
        props = tool.operator_properties("batoms.apply_cell")
        layout.prop(props, "a")
        layout.prop(props, "b")
        layout.prop(props, "c")


class BatomsTransform(WorkSpaceTool):
    bl_space_type = "VIEW_3D"
    bl_context_mode = "OBJECT"  # "OBJECT" or "EDIT_MESH"

    bl_idname = "batoms.apply_transform"
    bl_label = "Transform"
    bl_description = "Apply new transform parameters."
    # bl_icon = "ops.generic.select_circle"
    bl_icon = "ops.transform.shear"
    bl_widget = None

    bl_keymap = (
        (
            "batoms.apply_transform",
            {"type": "RIGHTMOUSE", "value": "PRESS"},
            {"properties": []},
        ),
        # ("batoms.apply_transform", {"type": "LEFTMOUSE", "value": "PRESS", "ctrl": True},
        #  {"properties": []}),
    )

    def draw_settings(context, layout, tool):
        props = tool.operator_properties("batoms.apply_transform")
        layout.prop(props, "a")
        layout.prop(props, "b")
        layout.prop(props, "c")


class BatomsBoundary(WorkSpaceTool):
    bl_space_type = "VIEW_3D"
    bl_context_mode = "OBJECT"  # "OBJECT" or "EDIT_MESH"

    bl_idname = "batoms.apply_boundary"
    bl_label = "Boundary"
    bl_description = "Apply new Boundary parameters."
    # bl_icon = "ops.generic.select_circle"
    bl_icon = "ops.transform.shrink_fatten"
    bl_widget = None

    bl_keymap = (
        (
            "batoms.apply_boundary",
            {"type": "RIGHTMOUSE", "value": "PRESS"},
            {"properties": []},
        ),
        # ("batoms.apply_boundary", {"type": "LEFTMOUSE", "value": "PRESS", "ctrl": True},
        #  {"properties": []}),
    )

    def draw_settings(context, layout, tool):
        props = tool.operator_properties("batoms.apply_boundary")
        layout.prop(props, "a")
        layout.prop(props, "b")
        layout.prop(props, "c")


class AddSurface(WorkSpaceTool):
    bl_space_type = "VIEW_3D"
    bl_context_mode = "OBJECT"

    bl_idname = "batoms.surface_add"
    bl_label = "Add surface"
    bl_description = "Create an surface structure from a bulk."
    bl_icon = "ops.transform.translate"

    def draw_settings(context, layout, tool):
        props = tool.operator_properties("batoms.surface_add")
        layout.prop(props, "label")
        layout.prop(props, "indices")
        layout.prop(props, "size")
        layout.prop(props, "vacuum")
        layout.prop(props, "nlayer")
        layout.prop(props, "termination")


class MoleculeEditElement(WorkSpaceTool):
    bl_space_type = "VIEW_3D"
    bl_context_mode = "EDIT_MESH"  # "OBJECT" or "EDIT_MESH"

    bl_idname = "batoms_tool.molecule_edit_atom"
    bl_label = "Edit atoms"
    bl_description = "Create an molecule structure from a database."
    # bl_icon = "ops.generic.select_circle"
    bl_icon = "batoms_molecule_edit_atom"
    bl_widget = None

    bl_keymap = (
        (
            "batoms.molecule_edit_atom",
            {"type": "RIGHTMOUSE", "value": "PRESS"},
            {"properties": []},
        ),
        (
            "batoms.molecule_edit_atom",
            {"type": "LEFTMOUSE", "value": "PRESS", "ctrl": True},
            {"properties": []},
        ),
    )

    def draw_settings(context, layout, tool):
        props = tool.operator_properties("batoms.molecule_edit_atom")
        layout.prop(props, "element")
        layout.prop(props, "bond_order")


class MolecueEditBond(WorkSpaceTool):
    bl_space_type = "VIEW_3D"
    bl_context_mode = "EDIT_MESH"  # "OBJECT" or "EDIT_MESH"

    bl_idname = "batoms_tool.molecule_edit_bond"
    bl_label = "Edit bond"
    bl_description = "Create an molecule structure from a database."
    # bl_icon = "ops.generic.select_circle"
    bl_icon = "batoms_molecule_edit_atom"
    bl_widget = None

    def draw_settings(context, layout, tool):
        props = tool.operator_properties("batoms.molecule_edit_bond")
        layout.prop(props, "order")
        layout.prop(props, "rotate")
        layout.prop(props, "hydrogen")



from setuptools import setup, find_packages

setup(
    name="batoms",
    version="2.4.0.1",
    packages=find_packages(),
    license="GPL",
    description="A Python package for creating, editing and rendering atoms and molecules structures using Blender.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Beautiful Atoms Team",
    author_email="xing.wang@gmail.com",
    url="https://github.com/beautiful-atoms/beautiful-atoms",
    install_requires=[
        "bpy",
        "ase",
    ],
    package_data={
        "batoms.asset.libraries": ["*.blend"],
    },
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.10, <4",
    entry_points={},
)



import bpy
from batoms import Batoms
from batoms.render import Camera
from ase.build import fcc111

try:
    from _common_helpers import has_display, set_cycles_res

    use_cycles = not has_display()
except ImportError:
    use_cycles = False

extras = dict(engine="cycles") if use_cycles else {}


def test_camera():
    """ """
    bpy.ops.batoms.delete()
    camera = Camera(label="h2o", name="1")
    assert isinstance(camera, Camera)
    # properties
    camera.type = "PERSP"
    assert camera.type == "PERSP"
    camera.lens = 50
    assert camera.lens == 50
    camera.target = [5, 0, 0]


def test_camera_ortho():
    bpy.ops.batoms.delete()
    au111 = fcc111("Au", (4, 4, 4), vacuum=0)
    au111 = Batoms("au111", from_ase=au111)
    au111.cell[2, 2] += 10
    # au111.draw_cell()
    if use_cycles:
        set_cycles_res(au111)
    au111.get_image([1, 0, 0], output="camera_ortho.png", **extras)
    au111.get_image(
        [0, 0, 1], center=au111.positions[-1], output="camera-center.png", **extras
    )


def test_camera_persp():
    bpy.ops.batoms.delete()
    au111 = fcc111("Au", (4, 10, 4), vacuum=0)
    au111 = Batoms("au111", from_ase=au111)
    au111.cell[2, 2] += 10
    # au111.draw_cell()
    au111.render.camera.type = "PERSP"
    if use_cycles:
        set_cycles_res(au111)
    au111.get_image([1, 0, 0], output="camera-persp.png", **extras)


if __name__ == "__main__":
    test_camera()
    test_camera_ortho()
    test_camera_persp()
    print("\n camera: All pass! \n")



def test_slicebatoms(ch4):
    """Setting sliced Batoms"""
    import numpy as np

    # one atoms
    ch4.model_style = 1
    assert ch4[0].species == "C"
    ch4[1].scale = 1
    assert ch4[1].scale == 1
    ch4[1].show = 0
    assert ch4[1].show == 0
    #     ch4[1].species = "Cl"
    #     assert ch4[1].species == "Cl"
    assert np.isclose(
        ch4[1].position, np.array([0.62911803, 0.62911803, 0.62911803])
    ).all()
    # > one atoms
    assert ch4[0:2].species[0] == "C"
    print(ch4[:].scale)
    assert np.isclose(ch4[0:2].scale, np.array([0.4, 1])).all()
    ch4[0:2].scale = 2
    assert np.isclose(ch4[0:2].scale, np.array([2, 2])).all()



"""
Molecular surface:
1) van der Waals surface
2) Accessible surface area (SAS)
3) Solvent-excluded surface (SES) or Connolly surface
"""
import bpy
from batoms.base.collection import Setting
import logging

logger = logging.getLogger(__name__)


class MolecularSurfaceSettings(Setting):
    def __init__(
        self,
        label,
        parent,
        probe=1.4,
        resolution=0.5,
    ) -> None:
        """MS object
        The MS object store the MS information.

        Parameters:

        label: str
            The label define the batoms object that
            a Setting belong to.
        probe: float
            probe radius, in [0,inf], default is 1.4
        """
        Setting.__init__(self, label, coll_name=label)
        self.label = label
        self.name = "Bmolecularsurface"
        self.probe = probe
        self.parent = parent
        self.sas_name = "%s_sas" % self.label
        self.ses_name = "%s_ses" % self.label
        self.resolution = resolution
        if len(self) == 0:
            self["1"] = {"select": "all"}

    def add(self, name, datas={}):
        self[name] = datas

    def set_collection(self, label):
        """ """
        if not bpy.data.collections.get(label):
            coll = bpy.data.collections.new(label)
            self.parent.batoms.coll.children.link(coll)
            coll.batoms.type = "MS"
            coll.batoms.label = label

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "name  type select  probe   resolution    color  \n"
        for ms in self.bpy_setting:
            s += "{:4s}   {:6s}   {:6s}   {:1.3f}  {:1.3f} ".format(
                ms.name, ms.type, ms.select, ms.probe, ms.resolution
            )
            s += "[{:1.1f}  {:1.1f}  {:1.1f}  {:1.1f}] \n".format(
                ms.color[0], ms.color[1], ms.color[2], ms.color[3]
            )
        s += "-" * 60 + "\n"
        return s



import bpy
from bpy.types import Panel
from bpy.props import (
    FloatProperty,
    EnumProperty,
)
from batoms.render.render import Render
from batoms.gui.utils import get_attr, set_attr


class Render_PT_prepare(Panel):
    bl_label = "Render"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"DEFAULT_CLOSED"}
    bl_category = "Batoms"
    bl_idname = "RENDER_PT_Tools"

    def draw(self, context):
        layout = self.layout
        render = context.scene.batoms.render

        layout = self.layout
        # row = col.row(align=True)
        layout.operator("batoms.render_add")
        layout.label(text="Viewport")
        row = layout.row()
        row.prop(render, "viewport_x")
        row.prop(render, "viewport_y")
        row.prop(render, "viewport_z")
        layout.separator()
        # col = layout.column()
        layout.label(text="Camera")
        layout.label(text="Type")
        layout.prop(render, "camera_type", expand=True)
        if render.camera_type == "ORTHO":
            layout.prop(render, "ortho_scale")
        else:
            layout.prop(render, "camera_lens")
            layout.prop(render, "distance")
        # layout.prop(render, "resolution")
        layout.separator()
        layout.label(text="Light")
        row = layout.row()
        row.prop(render, "light_direction_x")
        row.prop(render, "light_direction_y")
        row.prop(render, "light_direction_z")
        layout.prop(render, "energy")


# ---------------------------------------------------


def modify_render_attr(context, key, value):
    from batoms.batoms import Batoms

    if context.object and context.object.batoms.type != "OTHER":
        batoms = Batoms(label=context.object.batoms.label)
        setattr(batoms.render, key, value)
        if key not in ["distance"]:
            batoms.render.init()
        context.space_data.region_3d.view_perspective = "CAMERA"


def set_light_attr(key, value):
    """Set light attribute

    We need to use Render.

    Args:
        key (_type_): _description_
        value (_type_): _description_
    """
    from batoms.batoms import Batoms

    if bpy.context.object and bpy.context.object.batoms.type != "OTHER":
        batoms = Batoms(label=bpy.context.object.batoms.label)
        setattr(batoms.render.lights["Default"], key, value)
        batoms.render.init()
        bpy.context.space_data.region_3d.view_perspective = "CAMERA"


def set_camera_attr(key, value):
    """Set Camera attribute

    we need to use Batoms

    Args:
        key (_type_): _description_
        value (_type_): _description_
    """
    from batoms.batoms import Batoms

    if bpy.context.object and bpy.context.object.batoms.type != "OTHER":
        batoms = Batoms(label=bpy.context.object.batoms.label)
        if key == "ortho_scale":
            batoms.render.camera.set_ortho_scale(value)
        elif key in ["lens"]:
            setattr(batoms.render.camera, key, value)
        else:
            setattr(batoms.render.camera, key, value)
            batoms.render.init()
        bpy.context.space_data.region_3d.view_perspective = "CAMERA"


def get_active_render_collection():
    """Get the collection of the active Batoms

    When get the attribute of Batoms object,
    if the attribute if saved in the Batoms.coll.batoms,
    we only need to read data form the colleciton,
    it is faster than get data from the Batoms itself.

    Returns:
        bpy.type.collection: _description_
    """
    if "batoms_render" in bpy.data.collections:
        return bpy.data.collections["batoms_render"].Brender
    return None


def get_active_render():
    if "batoms_render" in bpy.data.collections:
        render = Render(label="batoms")
        return render
    return None


def get_default_camera():
    if "batoms_camera_Default" in bpy.data.objects:
        return bpy.data.objects["batoms_camera_Default"]
    return None


def get_default_camera_data():
    if "batoms_camera_Default" in bpy.data.objects:
        return bpy.data.objects["batoms_camera_Default"].data
    return None


def get_default_light():
    if "batoms_light_Default" in bpy.data.objects:
        return bpy.data.objects["batoms_light_Default"]
    return None


def get_default_light_dat():
    if "batoms_light_Default" in bpy.data.objects:
        return bpy.data.objects["batoms_light_Default"].data
    return None


def get_viewport(i):
    """Helper function to easily get cell-property."""

    def getter(self):
        render = get_active_render_collection()
        if render is not None:
            return render.viewport[i]
        else:
            return 0

    return getter


def set_viewport(i):
    """Helper function to easily get cell-property."""

    def setter(self, value):
        render = get_active_render_collection()
        if render is not None:
            viewport = render.viewport
            viewport[i] = value
            render = get_active_render()
            render.viewport = viewport

    return setter


def get_camera_type(self):
    camera = get_default_camera()
    if camera is not None:
        if camera.data.type == "ORTHO":
            return 0
        else:
            return 1
    else:
        return 0


def set_camera_type(self, value):
    items = self.bl_rna.properties["camera_type"].enum_items
    item = items[value]
    type = item.identifier
    # print(type)
    self["type"] = type
    set_camera_attr("type", type)


def get_distance(self):
    render = get_active_render_collection()
    if render is not None:
        return render.distance
    else:
        return 10


def set_distance(self, value):
    self["distance"] = value
    render = get_active_render()
    render.distance = value


def get_light_direction(i):
    """Helper function to easily get cell-property."""

    def getter(self):
        light = get_default_light()
        if light is not None:
            return light.Blight.direction[i]
        else:
            return 0

    return getter


def set_light_direction(i):
    """Helper function to easily set light-property."""

    def setter(self, value):
        light = get_default_light()
        if light is not None:
            light_direction = light.Blight.direction
            light_direction[i] = value
            set_light_attr("direction", light_direction)

    return setter


class RenderProperties(bpy.types.PropertyGroup):
    """ """

    viewport_x: FloatProperty(
        name="viewport_x",
        default=0,
        soft_min=-5,
        soft_max=5,
        description="Miller viewport for the render",
        get=get_viewport(0),
        set=set_viewport(0),
    )

    viewport_y: FloatProperty(
        name="viewport_y",
        default=0,
        soft_min=-5,
        soft_max=5,
        description="Miller viewport for the render",
        get=get_viewport(1),
        set=set_viewport(1),
    )

    viewport_z: FloatProperty(
        name="viewport_z",
        default=1,
        soft_min=-5,
        soft_max=5,
        description="Miller viewport for the render",
        get=get_viewport(2),
        set=set_viewport(2),
    )

    camera_type: EnumProperty(
        name="type",
        description="camera type",
        items=(("ORTHO", "ORTHO", "", 0), ("PERSP", "PERSP", "", 1)),
        default=0,
        get=get_camera_type,
        set=set_camera_type,
    )

    # resolution: IntVectorProperty(
    # name="resolution", size=2, default=(1000, 1000),
    # soft_min = 100, soft_max = 2000,
    # description="Miller resolution for the render", update=Callback_modify_resolution)

    ortho_scale: FloatProperty(
        name="ortho_scale",
        default=10,
        soft_min=1,
        soft_max=100,
        description="ortho_scale",
        get=get_attr("ortho_scale", get_default_camera_data),
        set=set_attr("ortho_scale", set_camera_attr),
    )

    light_direction_x: FloatProperty(
        name="light_direction_x",
        default=0,
        soft_min=-5,
        soft_max=5,
        description="Light direction for the render",
        get=get_light_direction(0),
        set=set_light_direction(0),
    )

    light_direction_y: FloatProperty(
        name="light_direction_y",
        default=0,
        soft_min=-5,
        soft_max=5,
        description="Light direction for the render",
        get=get_light_direction(1),
        set=set_light_direction(1),
    )

    light_direction_z: FloatProperty(
        name="light_direction_z",
        default=1,
        soft_min=-5,
        soft_max=5,
        description="Light direction for the render",
        get=get_light_direction(2),
        set=set_light_direction(2),
    )

    energy: FloatProperty(
        name="energy",
        default=10,
        soft_min=1,
        soft_max=100,
        description="energy",
        get=get_attr("energy", get_default_light_dat),
        set=set_attr("energy", set_light_attr),
    )

    distance: FloatProperty(
        name="distance",
        default=3,
        description="distance from origin",
        get=get_distance,
        set=set_distance,
    )

    camera_lens: FloatProperty(
        name="lens",
        default=100,
        soft_min=1,
        soft_max=100,
        get=get_attr("lens", get_default_camera_data),
        set=set_attr("lens", set_camera_attr),
    )



"""
This module defines abstract Operator classes.
"""

from bpy.types import Operator
from batoms import Batoms


class OperatorBatoms(Operator):
    """ """

    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj:
            return obj.batoms.type != "OTHER"
        else:
            return False

    def execute(self, context):
        obj = context.object
        Batoms(label=obj.batoms.label)
        return {"FINISHED"}


class OperatorBatomsEdit(OperatorBatoms):
    """ """

    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj:
            return obj.batoms.type != "OTHER" and context.mode in {"EDIT_MESH"}
        else:
            return False



"""Definition of the plugin class.

This module defines the plugin object in the Batoms package.

"""

import bpy
from time import time
from batoms.plugins.base import PluginObject
from .setting import TemplateSettings
import logging

logger = logging.getLogger(__name__)


class Template(PluginObject):
    def __init__(
        self,
        label,
        batoms=None,
    ):
        """Template Class

        Args:
            label (string): _description_.
            batoms (Batoms, optional): _description_. Defaults to None.
        """
        #
        self.batoms = batoms
        self.label = label
        PluginObject.__init__(self, label, "template", batoms)
        self.settings = TemplateSettings(self.label, parent=self)
        self.settings.bpy_data.active = True

    def build_materials(
        self,
        name,
        color,
        node_inputs=None,
        material_style="default",
        backface_culling=False,
    ):
        """Create materials

        Args:
            name (string): _description_
            color (array, (1x4)): _description_
            node_inputs (dict, optional): _description_. Defaults to None.
            material_style (string, optional): _description_. Defaults to 'default'.

        Returns:
            bpy.type.material: _description_
        """
        from batoms.material import create_material

        if name in bpy.data.materials:
            mat = bpy.data.materials.get(name)
            bpy.data.materials.remove(mat, do_unlink=True)
        mat = create_material(
            name,
            color=color,
            node_inputs=node_inputs,
            material_style=material_style,
            backface_culling=backface_culling,
        )
        return mat

    def draw(self):
        from batoms.utils.butils import clean_coll_object_by_type

        # delete old object
        clean_coll_object_by_type(self.batoms.coll, "TEMPLATE")
        for setting in self.settings.bpy_setting:
            self.draw_plugin(setting)

    def draw_plugin(self, setting):
        """ """
        tstart = time()
        logger.debug("Draw Template: %s" % (time() - tstart))

    @property
    def objs(self):
        objs = {}
        for setting in self.settings.bpy_setting:
            ms_name = "%s_%s" % (self.label, setting.name)
            obj = bpy.data.objects.get(ms_name)
            objs[setting.name] = obj
        return objs

    @property
    def mat(self):
        return bpy.data.materials.get(self.name)

    def as_dict(self):
        """ """
        data = {}
        data["settings"] = self.settings.as_dict()
        data.update(self.settings.bpy_data.as_dict())
        return data



"""
Get and set attribute
"""
import numpy as np
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)


def get_mesh_attribute_bmesh(obj, key, index=None):
    """Get the attribute of mesh by name usinb bmesh method.
    When use this function:
    1) Edit mode, use bmesh.
    2) For STRING attribute

    Args:
        key (string): name of the attribute
        index (bool, int): index of the data, used to get singe attribute value

    Raises:
        KeyError: _description_

    Returns:
        array: _description_
    """
    import bmesh
    from batoms.utils.butils import get_att_length, get_bmesh_layer, get_bmesh_domain
    from batoms.utils import type_blender_to_py

    # get the mesh
    me = obj.data
    # get attribute data type for the attribute
    att = me.attributes.get(key)
    if att is None:
        raise KeyError("{} is not exist.".format(key))
    # get type and domain
    dtype = att.data_type
    # get attribute length based on domain
    n = get_att_length(obj.data, att)
    # check mode
    if obj.mode == "EDIT":
        bm = bmesh.from_edit_mesh(obj.data)
        domain = get_bmesh_domain(bm, att)
        domain.ensure_lookup_table()
        layer = get_bmesh_layer(domain, key, dtype)
        if index is not None:
            return np.array([domain[index][layer]])
        # init attribute array
        attribute = np.zeros(n, dtype=type_blender_to_py(dtype, str="U20"))
        for i in range(n):
            # print(domain[i][layer])
            attribute[i] = domain[i][layer]
    else:
        bm = bmesh.new()
        bm.from_mesh(me)
        domain.ensure_lookup_table()
        domain = get_bmesh_domain(bm, att)
        if index is not None:
            return np.array([domain[index][layer]])
        n = len(domain)
        for i in range(n):
            attribute[i] = domain[i][layer]
        bm.free()
    return attribute


def get_mesh_attribute(obj, key, index=None):
    """Get the attribute of mesh by name usinb bmesh method.

    When use this function:
    1) Object mode, read attribute directly
    2) For BOOLEAN attribute, because bmesh does not support BOOLEAN.

    Args:
        key (string): name of the attribute
        index (bool, int): index of the data, used to get singe attribute value

    Raises:
        KeyError: _description_

    Returns:
        array: _description_
    """
    from batoms.utils.butils import get_att_length
    from batoms.utils import type_blender_to_py

    # get the mesh
    me = obj.data
    # get attribute data type for the attribute
    att = me.attributes.get(key)
    # print(f"Attribute: {att.data_type} {att.domain}")
    if att is None:
        raise KeyError("{} is not exist.".format(key))
    dtype = att.data_type
    # get single attribute value
    if index is not None:
        return np.array([att.data[index].value])
    else:
        # get attribute length based on domain
        n = get_att_length(obj.data, att)
        # init
        attribute = np.zeros(n, dtype=type_blender_to_py(dtype, str="U20"))
        if dtype == "STRING":
            for i in range(n):
                attribute[i] = att.data[i].value
        elif dtype in ["INT", "FLOAT", "BOOLEAN"]:
            att.data.foreach_get("value", attribute)
        elif dtype == "FLOAT2":
            attribute = np.zeros(n * 2, dtype="float")
            att.data.foreach_get("vector", attribute)
            attribute = attribute.reshape(n, 2)
        elif dtype == "FLOAT_VECTOR":
            attribute = np.zeros(n * 3, dtype="float")
            att.data.foreach_get("vector", attribute)
            attribute = attribute.reshape(n, 3)
        elif dtype in ["FLOAT_COLOR"]:
            attribute = np.zeros(n * 4, dtype="float")
            att.data.foreach_get("color", attribute)
            attribute = attribute.reshape(n, 4)
        elif dtype in ["QUATERNION"]:
            attribute = np.zeros(n * 4, dtype="float")
            att.data.foreach_get("value", attribute)
            attribute = attribute.reshape(n, 4)
        else:
            raise KeyError("Attribute type: %s is not support." % dtype)
        attribute = np.array(attribute)
    return attribute


def set_mesh_attribute_bmesh(obj, key, value, index=None):
    """Set mesh attribute using bmesh method.

    Args:
        obj (bpy.type.object): obj
        key (str): name of the attribute
        value (np.array): value of the attribute
        index (bool, int): index of the data, used to set singe attribute value
    """
    import bmesh
    from batoms.utils.butils import get_bmesh_domain, get_bmesh_layer

    me = obj.data
    # get attribute type
    att = me.attributes.get(key)
    dtype = att.data_type
    if dtype == "STRING":
        value = np.char.encode(value)
    # check object mode
    if obj.mode == "EDIT":
        bm = bmesh.from_edit_mesh(me)
        domain = get_bmesh_domain(bm, att)
        domain.ensure_lookup_table()
        layer = get_bmesh_layer(domain, key, dtype)
        if index is not None:
            domain[index][layer] = value
        else:
            n = len(domain)
            for i in range(n):
                domain[i][layer] = value[i]
        bmesh.update_edit_mesh(me)
    else:
        bm = bmesh.new()
        bm.from_mesh(me)
        domain = get_bmesh_domain(bm, att)
        domain.ensure_lookup_table()
        layer = get_bmesh_layer(domain, key, dtype)
        if index is not None:
            domain[index][layer] = value
        else:
            n = len(domain)
            for i in range(n):
                domain[i][layer] = value[i]
        bm.to_mesh(me)
        bm.free()


def set_mesh_attribute(obj, key, value, index=None):
    """Set mesh attribute using bmesh method

    Args:
        obj (bpy.type.object): obj
        key (str): name of the attribute
        value (np.array): value of the attribute
        index (bool, int): index of the data, used to set singe attribute value
    """
    from batoms.utils.butils import get_att_length

    me = obj.data
    # get attribute
    att = me.attributes.get(key)
    # print(f"Attribute: {att.data_type} {att.domain}")
    if index is not None:
        att.data[index].value = value[0]
    else:
        # get attribute domain and length
        n = get_att_length(obj.data, att)
        if att.data_type == "STRING":
            for j in range(n):
                att.data[j].value = value[j]
        elif att.data_type == "FLOAT2":
            value = value.reshape((n * 2, 1))
            att.data.foreach_set("vector", value)
        elif att.data_type == "FLOAT_VECTOR":
            value = value.reshape((n * 3, 1))
            att.data.foreach_set("vector", value)
        elif att.data_type in ["FLOAT_COLOR"]:
            value = value.reshape((n * 4, 1))
            att.data.foreach_set("color", value)
        elif att.data_type in ["QUATERNION"]:
            value = value.reshape((n * 4, 1))
            att.data.foreach_set("value", value)
        else:
            att.data.foreach_set("value", value)



import bpy
import pytest
from ase.build import bulk
from batoms import Batoms
from batoms.bio.bio import read
import os

try:
    from _common_helpers import has_display, set_cycles_res

    use_cycles = not has_display()
except ImportError:
    use_cycles = False

extras = dict(engine="cycles") if use_cycles else {}

skip_test = bool(os.environ.get("NOTEST_CUBE", 0))


def test_lattice_plane():
    bpy.ops.batoms.delete()
    au = bulk("Au", cubic=True)
    au = Batoms("au", from_ase=au)
    au.lattice_plane.settings[(1, 1, 1)] = {"distance": 3, "scale": 1.1}
    au.lattice_plane.settings[(1, 1, 0)] = {"distance": 3, "color": [0.8, 0.1, 0, 0.8]}
    print(au.lattice_plane.settings)
    au.lattice_plane.draw()
    # au.draw_cell()
    if use_cycles:
        set_cycles_res(au)
    au.get_image([0, 0, 1], output="plane-lattice-plane.png", **extras)


def test_boundary():
    # skip due to cubefile
    if skip_test:
        pytest.skip("Skip tests on cube files since $NOTEST_CUBE provided.")
    bpy.ops.batoms.delete()
    h2o = read("../tests/datas/h2o-homo.cube")
    h2o.lattice_plane.settings[(0, 0, 1)] = {"distance": 6, "boundary": True}
    h2o.lattice_plane.settings[(0, 0, -1)] = {"distance": -5, "boundary": True}
    h2o.lattice_plane.draw()
    if use_cycles:
        set_cycles_res(h2o)
    h2o.get_image([1, 0, 0], output="plane-boundary.png", **extras)


def test_lattice_plane_ops():
    bpy.ops.batoms.delete()
    bpy.ops.batoms.bulk_add(formula="Au")
    au = Batoms("Au")
    bpy.context.view_layer.objects.active = au.obj
    bpy.ops.plane.lattice_plane_add(indices=(1, 1, 1))
    au.lattice_plane.settings["1-1-1"].distance = 3.0
    bpy.ops.plane.lattice_plane_add(indices=(1, 0, 0))
    assert len(au.lattice_plane.settings) == 2
    bpy.ops.plane.lattice_plane_draw()
    bpy.ops.plane.lattice_plane_remove(name="1-0-0")
    assert len(au.lattice_plane.settings) == 1
    print(au.lattice_plane.settings)
    bpy.ops.plane.lattice_plane_draw()


def test_gui():
    """latticeplane panel"""
    from batoms.batoms import Batoms

    bpy.ops.batoms.delete()
    bpy.ops.batoms.bulk_add(formula="Au", cubic=True)
    au = Batoms("Au")
    au.obj.select_set(True)
    assert bpy.context.scene.Blatticeplane.show is True
    bpy.context.scene.Blatticeplane.show = False
    assert au.lattice_plane.show is False


def test_gui_uilist():
    """latticeplane panel"""
    from batoms.batoms import Batoms

    bpy.ops.batoms.delete()
    bpy.ops.batoms.bulk_add(formula="Au", cubic=True)
    au = Batoms("Au")
    au.obj.select_set(True)
    assert au.coll.Blatticeplane.ui_list_index == 0
    bpy.ops.plane.lattice_plane_add(indices=(1, 1, 1))
    bpy.ops.plane.lattice_plane_add(indices=(1, 0, 0))
    assert au.coll.Blatticeplane.ui_list_index == 1


if __name__ == "__main__":
    test_lattice_plane()
    test_boundary()
    print("\n Lattice plane: All pass! \n")



import bpy
from batoms import Batoms
import numpy as np


def test_boundary():
    bpy.ops.batoms.delete()
    o = Batoms("o", ["O"], [[1, 1, 1]], pbc=True, cell=[2, 2, 2])
    o.boundary = [1, 0, 0]
    assert len(o.boundary) == 2


def test_boundary_scale(au):
    au.scale = 0.5
    au.boundary = [1, 1, 1]
    assert np.allclose(au.boundary.get_attribute("scale")[0], 0.5)
    # if bpy.app.version_string >= '3.2.0':
    # au.scale = 1.0
    # assert np.allclose(au.boundary.get_attribute('scale')[0], 1)


def test_boundary_off_origin(au):
    au.boundary = [1, 1, 0]
    assert np.allclose(au.boundary.positions[0], np.array([0, -2.03999996, 2.03999996]))
    # repeat
    au.translate([0, 0, 2])
    au.boundary = [1, 1, 0]
    assert np.allclose(
        au.boundary.global_positions[0], np.array([0, -2.03999996, 4.03999996])
    )


def test_boundary_oxide(tio2):
    tio2.boundary = 0.01
    assert len(tio2.boundary.obj.data.vertices) == 9
    assert np.allclose(tio2.boundary.positions[0], np.array([0, 0, 2.969203]))


def test_boundary_animation():
    from ase.io import read

    bpy.ops.batoms.delete()
    tio2 = read("../tests/datas/tio2_10.xyz", ":")
    tio2 = Batoms("tio2", from_ase=tio2, load_trajectory=True)
    tio2.boundary = 0.01
    assert len(tio2.boundary.obj.data.vertices) == 9


def test_boundary_reload(au):
    """save to blend file and reload"""
    import os

    au.boundary = 0.1
    cwd = os.getcwd()
    filepath = os.path.join(cwd, "test.blend")
    bpy.ops.wm.save_as_mainfile(filepath=filepath)
    bpy.ops.batoms.delete()
    bpy.ops.wm.open_mainfile(filepath=filepath)
    au = Batoms("au")
    np.isclose(au.boundary[0, 0], -1)
    assert len(au.boundary) == 10



"""Definition of the highlight class.

This module defines the highlight object in the Batoms package.

"""

import bpy
import numpy as np
from batoms.base.object import ObjectGN
from batoms.plugins.base import PluginObject
from .setting import HighlightSettings
from batoms.utils.butils import get_node_by_name
from batoms.utils import string2Number
import logging

logger = logging.getLogger(__name__)

default_attributes = [
    {"name": "atom_index", "data_type": "INT"},
    {"name": "select_index", "data_type": "INT"},
    {"name": "select", "data_type": "STRING"},
    {"name": "show", "data_type": "INT"},
    {"name": "scale", "data_type": "FLOAT"},
]

default_GroupInput = [
    ["atom_index", "NodeSocketInt"],
    ["select_index", "NodeSocketInt"],
    ["show", "NodeSocketBool"],
    ["scale", "NodeSocketFloat"],
]

default_highlight_datas = {
    "select_index": np.ones(0, dtype=int),
    "atom_index": np.ones(0, dtype=int),
    "centers": np.zeros((1, 0, 3)),
    "show": np.zeros(0, dtype=int),
    "scale": np.ones(0, dtype=float),
}


class Highlight(ObjectGN, PluginObject):
    def __init__(
        self,
        label=None,
        highlight_datas=None,
        location=np.array([0, 0, 0]),
        batoms=None,
    ):
        """Highlight Class

        Args:
            label (_type_, optional): _description_. Defaults to None.
            location (_type_, optional): _description_. Defaults to np.array([0, 0, 0]).
            batoms (_type_, optional): _description_. Defaults to None.
        """
        #
        self.batoms = batoms
        self.label = label
        name = "highlight"
        ObjectGN.__init__(self, label, name)
        PluginObject.__init__(self, label, "highlight", batoms)
        flag = self.load()
        if not flag:
            if highlight_datas is None:
                highlight_datas = default_highlight_datas
            self.build_object(highlight_datas)
            self.settings = HighlightSettings(self.label, batoms=batoms, parent=self)
        else:
            self.settings = HighlightSettings(self.label, batoms=batoms, parent=self)
        self.settings.bpy_data.active = True

    def build_materials(self, name, color, node_inputs=None, material_style="default"):
        """ """
        from batoms.material import create_material

        if name in bpy.data.materials:
            mat = bpy.data.materials.get(name)
            bpy.data.materials.remove(mat, do_unlink=True)
        mat = create_material(
            name,
            color=color,
            node_inputs=node_inputs,
            material_style=material_style,
            backface_culling=False,
        )
        return mat

    def draw(self):
        """Draw highlight."""
        # delete old highlight
        highlight_datas = self.build_highlight()
        self.set_arrays(highlight_datas)

    def build_highlight(self):
        """Find cage base on radius range and boundary range
        Algorith:
            1) build a meshgrid
            2) calulated first neighbour distance by kdtree
            3) find the max distance (max radius sphere)
            4) remove meshgrid within this sphere,
            5) find next max radius, ..., untill radius < minCave
            6) remove all spheres contact with boundary
        Args:
            cell (_type_): _description_
            positions (_type_): _description_
            radius (_type_): _description_
            step (float, optional): _description_. Defaults to 1.0.
            boundary (list, optional): _description_. Defaults to [[0, 1], [0, 1], [0, 1]].

        Returns:
            _type_: _description_
        """
        arrays = self.batoms.arrays
        natom = 0
        for hl in self.settings.bpy_setting:
            indices = self.batoms.selects[hl.select].indices
            natom += len(indices)
        highlights = {
            "select_index": np.ones(natom, dtype=int),
            "atom_index": np.ones(natom, dtype=int),
            "centers": np.zeros((1, natom, 3)),
            "show": np.zeros(natom, dtype=int),
            "scale": np.ones(natom, dtype=float),
        }
        i = 0
        for hl in self.settings.bpy_setting:
            indices = self.batoms.selects[hl.select].indices
            n = len(indices)
            highlights["centers"][0][i : i + n] = arrays["positions"][indices]
            highlights["scale"][i : i + n] = arrays["scale"][indices] * hl.scale
            highlights["show"][i : i + n] = np.ones(len(indices))
            highlights["select_index"][i : i + n] = np.ones(
                len(indices)
            ) * string2Number(hl.name)
            highlights["atom_index"][i : i + n] = np.array(indices)
            i += n
        return highlights

    def build_object(self, highlight_datas, attributes={}):
        """Build the main Batoms object

        Args:
            label (str):
                Name of the object
            arrays (array):
                arrays of properties for each atoms
            location (list, optional):
                Location of the object. Defaults to [0, 0, 0].
        """
        if len(highlight_datas["centers"].shape) == 2:
            self._trajectory = {
                "centers": np.array([highlight_datas["centers"]]),
            }
            centers = highlight_datas["centers"]
        elif len(highlight_datas["centers"].shape) == 3:
            self._trajectory = {
                "centers": highlight_datas["centers"],
            }
            centers = highlight_datas["centers"][0]
        else:
            raise Exception("Shape of centers is wrong!")
        attributes.update(
            {
                "select_index": highlight_datas["select_index"],
                "show": highlight_datas["show"],
                "scale": highlight_datas["scale"],
            }
        )
        name = self.obj_name
        self.delete_obj(name)
        mesh = bpy.data.meshes.new(name)
        obj = bpy.data.objects.new(name, mesh)
        obj.data.from_pydata(centers, [], [])
        obj.batoms.type = "HIGHLIGHT"
        obj.batoms.label = self.batoms.label
        self.batoms.coll.objects.link(obj)
        # Add attributes
        for att in default_attributes:
            self.add_attribute(**att)
        # add cell object as its child
        obj.parent = self.batoms.obj
        self.set_attributes(attributes)
        self.init_geometry_node_modifier(default_GroupInput)
        self.build_geometry_node()

    def build_geometry_node(self):
        """Geometry node for instancing sphere on vertices!"""
        nodes = self.gn_node_group.nodes
        links = self.gn_node_group.links
        GroupInput = nodes[0]
        GroupOutput = nodes[1]
        # print(gn.name)
        JoinGeometry = get_node_by_name(
            nodes, "%s_JoinGeometry" % self.label, "GeometryNodeJoinGeometry"
        )
        SeparateGeometry = get_node_by_name(
            nodes, "%s_SeparateGeometry" % self.label, "GeometryNodeSeparateGeometry"
        )
        links.new(GroupInput.outputs["Geometry"], SeparateGeometry.inputs["Geometry"])
        links.new(GroupInput.outputs[2], SeparateGeometry.inputs["Selection"])
        links.new(SeparateGeometry.outputs[0], JoinGeometry.inputs["Geometry"])
        links.new(JoinGeometry.outputs["Geometry"], GroupOutput.inputs["Geometry"])
        #
        # transform postions of batoms to boundary
        ObjectBatoms = get_node_by_name(
            nodes, "%s_ObjectBatoms" % self.label, "GeometryNodeObjectInfo"
        )
        ObjectBatoms.inputs["Object"].default_value = self.batoms.obj
        PositionBatoms = get_node_by_name(
            nodes, "%s_PositionBatoms" % (self.label), "GeometryNodeInputPosition"
        )
        TransferBatoms = get_node_by_name(
            nodes, "%s_TransferBatoms" % (self.label), "GeometryNodeSampleIndex"
        )
        TransferBatoms.data_type = "FLOAT_VECTOR"
        links.new(ObjectBatoms.outputs["Geometry"], TransferBatoms.inputs[0])
        links.new(PositionBatoms.outputs["Position"], TransferBatoms.inputs[3])
        links.new(GroupInput.outputs[1], TransferBatoms.inputs["Index"])
        #
        # set positions
        SetPosition = get_node_by_name(
            nodes, "%s_SetPosition" % self.label, "GeometryNodeSetPosition"
        )
        links.new(GroupInput.outputs["Geometry"], SetPosition.inputs["Geometry"])
        links.new(TransferBatoms.outputs[0], SetPosition.inputs["Position"])

    def add_geometry_node(self, slname, instancer):
        """Add geometry node for each select

        Args:
            slname (str):
                Name of the select
            instancer (bpy.data.object):
                Object to be instanced
        """
        from batoms.utils.butils import compareNodeType

        nodes = self.gn_node_group.nodes
        links = self.gn_node_group.links
        GroupInput = nodes[0]
        SetPosition = get_node_by_name(nodes, "%s_SetPosition" % self.label)
        JoinGeometry = get_node_by_name(
            nodes, "%s_JoinGeometry" % self.label, "GeometryNodeJoinGeometry"
        )
        CompareSelect = get_node_by_name(
            nodes, "CompareFloats_%s_%s" % (self.label, slname), compareNodeType
        )
        CompareSelect.operation = "EQUAL"
        # CompareSelect.data_type = 'INT'
        CompareSelect.inputs[1].default_value = string2Number(slname)
        InstanceOnPoint = get_node_by_name(
            nodes,
            "InstanceOnPoint_%s_%s" % (self.label, slname),
            "GeometryNodeInstanceOnPoints",
        )
        ObjectInfo = get_node_by_name(
            nodes, "ObjectInfo_%s_%s" % (self.label, slname), "GeometryNodeObjectInfo"
        )
        ObjectInfo.inputs["Object"].default_value = instancer
        BoolShow = get_node_by_name(
            nodes,
            "BooleanMath_%s_%s_1" % (self.label, slname),
            "FunctionNodeBooleanMath",
        )
        #
        links.new(SetPosition.outputs["Geometry"], InstanceOnPoint.inputs["Points"])
        links.new(GroupInput.outputs[2], CompareSelect.inputs[0])
        links.new(GroupInput.outputs[3], BoolShow.inputs[0])
        links.new(GroupInput.outputs[4], InstanceOnPoint.inputs["Scale"])
        links.new(CompareSelect.outputs[0], BoolShow.inputs[1])
        links.new(BoolShow.outputs["Boolean"], InstanceOnPoint.inputs["Selection"])
        links.new(ObjectInfo.outputs["Geometry"], InstanceOnPoint.inputs["Instance"])
        links.new(InstanceOnPoint.outputs["Instances"], JoinGeometry.inputs["Geometry"])

    def update_geometry_node_instancer(self):
        """ """
        for sp in self.settings.bpy_setting:
            self.settings.build_instancer(sp.as_dict())

    def set_arrays(self, arrays):
        """ """
        # if len(arrays['positions']) == 0:
        #     return
        attributes = self.attributes
        print(arrays)
        # same length
        dnvert = len(arrays["select_index"]) - len(attributes["select_index"])
        if dnvert > 0:
            self.add_vertices_bmesh(dnvert)
        elif dnvert < 0:
            self.delete_vertices_bmesh(range(-dnvert))
        self.set_trajectory(arrays)
        self.set_attributes({"select_index": arrays["select_index"]})
        self.set_attributes({"scale": arrays["scale"]})
        self.set_attributes({"show": arrays["show"]})
        self.set_attributes({"atom_index": arrays["atom_index"]})
        self.update_mesh()
        self.update_geometry_node_instancer()

    def set_trajectory(self, frames=None, frame_start=0):
        if frames is None:
            frames = self._trajectory
        nframe = len(frames["centers"])
        if nframe == 0:
            return
        name = "%s_highlight" % (self.label)
        obj = self.obj
        self.set_shape_key(name, obj, frames["centers"], frame_start=frame_start)

    @property
    def objs(self):
        objs = {}
        name = "{}_{}".format(self.label, self.name)
        obj = bpy.data.objects.get(name)
        objs[name] = obj
        return objs

    @property
    def setting(self):
        from batoms.utils import deprecated

        """setting object."""
        deprecated(
            '"setting" will be deprecated in the furture, please use "settings".'
        )
        return self.settings

    def as_dict(self):
        """ """
        data = {}
        data["settings"] = self.settings.as_dict()
        data.update(self.settings.bpy_data.as_dict())
        return data



"""
"""
import bpy
from bpy_extras import view3d_utils
from bpy.props import (
    StringProperty,
)
from batoms import Batoms
from batoms.ops.base import OperatorBatoms
import blf
import numpy as np


def draw_callback_text(context, positions, texts):
    # get the context arguments
    n = len(positions)
    if n == 0:
        return
    region = context.region
    rv3d = context.region_data
    # coord = event.mouse_region_x, event.mouse_region_y
    font_id = 0  # XXX, need to find out how best to get this.
    for i in range(n):
        coord = np.mean(positions[i], axis=0)
        coord_2d = view3d_utils.location_3d_to_region_2d(
            region, rv3d, coord, default=None
        )
        # draw some text
        blf.position(font_id, coord_2d[0], coord_2d[1] + 10, 0)
        blf.size(font_id, 20, 72)
        blf.draw(font_id, texts[i])


class BatomsLabel(OperatorBatoms):
    """ """

    bl_idname = "batoms.label"
    bl_label = "Label"

    label: StringProperty(name="Label", default="Index", description="Label")

    select: StringProperty(name="Label", default="Index", description="Label")

    def execute(self, context):
        batoms = Batoms(label=context.object.batoms.label)
        batoms.show_label(self.label)
        bpy.context.view_layer.objects.active = batoms.obj
        return {"FINISHED"}



from .highlight import Highlight

from . import (
    bpy_data,
    ops,
    gui,
)

__all__ = ["Highlight"]

classes_bpy_data = [
    # internal data first
    bpy_data.HighlightSetting,
    bpy_data.Highlight,
    gui.HighlightProperties,
]

classes = [
    ops.HighlightAdd,
    ops.HighlightRemove,
    ops.HighlightDraw,
    gui.VIEW3D_PT_Batoms_highlight,
    gui.BATOMS_MT_highlight_context_menu,
    gui.BATOMS_UL_highlight,
    gui.BATOMS_PT_highlight,
]


def register_class():
    from bpy.types import Collection, Object, Scene
    from bpy.props import PointerProperty
    from bpy.utils import register_class

    for cls in classes_bpy_data:
        register_class(cls)
    # attach to blender internal data
    Collection.Bhighlight = PointerProperty(name="Bhighlight", type=bpy_data.Highlight)
    Object.Bhighlight = PointerProperty(name="Bhighlight", type=bpy_data.Highlight)
    Scene.Bhighlight = PointerProperty(type=gui.HighlightProperties)
    #
    for cls in classes:
        register_class(cls)


def unregister_class():
    from bpy.types import Collection, Object, Scene
    from bpy.utils import unregister_class

    del Collection.Bhighlight
    del Object.Bhighlight
    del Scene.Bhighlight
    for cls in reversed(classes):
        unregister_class(cls)
    for cls in reversed(classes_bpy_data):
        unregister_class(cls)



"""
Manipulate atoms and molecules with mouse interactively
based on force field methods.


Molecule: Rigid molecules, to constrain all internal
degrees of freedom using the RATTLE-type constraints of the
The FixBondLengths class to constrain all internal atomic distances.

Intermolecular: atomic radii.

"""

import bpy
import numpy as np
from batoms.utils.butils import get_selected_batoms, read_batoms_list
from batoms import Batoms
from batoms.data import covalent_radii
from bpy.props import (
    StringProperty,
    BoolProperty,
    FloatVectorProperty,
)
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)


def translate(selected_batoms, displacement):
    """ """
    from ase.geometry import get_distances

    displacement = displacement[:3] * 0.1
    # for ba
    batoms = Batoms(selected_batoms)
    atoms = batoms.atoms
    radii = covalent_radii[atoms.get_atomic_numbers()]
    n = len(atoms)
    batoms_list = read_batoms_list()
    flag = True
    for name in batoms_list:
        if name == selected_batoms:
            continue
        batoms1 = Batoms(name)
        batoms1.select = False
        atoms1 = batoms1.atoms
        radii1 = covalent_radii[atoms1.get_atomic_numbers()]
        n1 = len(atoms1)
        radii_mat = np.tile(radii, (n1, 1)).T
        radii1_mat = np.tile(radii1, (n, 1))
        radii_sum = radii_mat + radii1_mat
        p = atoms.positions + displacement
        p1 = atoms1.positions
        dvec, dmat = get_distances(p, p1)
        flag_mat = np.less(dmat, radii_sum)
        if flag_mat.any():
            flag = False
    if flag:
        batoms.translate(displacement)


def mouse2positions(delta, mat):
    """
    using rotation matrix
    """
    displacement = delta @ mat
    return displacement


class Rigid_Body_Operator(bpy.types.Operator):
    """Rigid body"""

    bl_idname = "object.rigid_body_operator"
    bl_label = "Rigi body translate Operator"

    mouse_position: FloatVectorProperty(size=4, default=(0, 0, 0, 0))
    previous: StringProperty()

    @property
    def selected_batoms(self):
        return get_selected_batoms()

    def modal(self, context, event):
        """ """
        if event.type in {"ESC"}:
            return {"CANCELLED"}
        elif event.type == "MIDDLEMOUSE":
            self.previous = "MIDDLEMOUSE"
            return {"PASS_THROUGH"}
        elif event.type == "MOUSEMOVE":
            mouse_position = np.array([event.mouse_x, event.mouse_y, 0, 0])
            if self.previous == "MIDDLEMOUSE":
                self.mouse_position = mouse_position
                self.previous = "MOUSEMOVE"
                return {"RUNNING_MODAL"}
            if len(self.selected_batoms) != 1:
                self.mouse_position = mouse_position
                return {"RUNNING_MODAL"}
            delta = mouse_position - self.mouse_position
            displacement = mouse2positions(
                delta, self.viewports_3D.spaces.active.region_3d.view_matrix
            )
            translate(self.selected_batoms[0], displacement)
            self.previous = "MOUSEMOVE"
            return {"RUNNING_MODAL"}
        else:
            mouse_position = np.array([event.mouse_x, event.mouse_y, 0, 0])
            self.mouse_position = mouse_position
            return {"PASS_THROUGH"}

    def invoke(self, context, event):
        """ """
        if context.object:
            for area in bpy.context.screen.areas:
                if area.type == "VIEW_3D":
                    self.viewports_3D = area
            self.mouse_position = np.array([event.mouse_x, event.mouse_y, 0, 0])
            context.window_manager.modal_handler_add(self)
            return {"RUNNING_MODAL"}
        else:
            self.report({"WARNING"}, "No active object, could not finish")
            return {"CANCELLED"}


class Rigid_Body_Modal_Panel(bpy.types.Panel):
    bl_idname = "rigid_body_modal_panel"
    bl_label = "Real Interactive"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "Real"
    bl_idname = "RIGIDBODY_PT_Modal"

    def draw(self, context):
        rbpanel = context.scene.rbpanel
        layout = self.layout
        box = layout.box()
        row = box.row()
        row.prop(rbpanel, "fix", expand=True)

        box = layout.box()
        row = box.row()
        row.operator("object.rigid_body_operator", text="Rigid Body", icon="FACESEL")


class RigidBodyProperties(bpy.types.PropertyGroup):
    @property
    def selected_batoms(self):
        return get_selected_batoms()

    def Callback_modify_fix(self, context):
        # clpanel = bpy.context.scene.clpanel
        # transform = clpanel.transform
        # modify_transform(self.selected_batoms, transform)
        pass

    def Callback_modify_cell(self, context):
        # clpanel = bpy.context.scene.clpanel
        # cell = [clpanel.cell_a, clpanel.cell_b, clpanel.cell_c]
        # modify_batoms_attr(self.selected_batoms, 'cell', cell)
        pass

    fix: BoolProperty(
        name="fix", default=True, description="fix", update=Callback_modify_fix
    )



import bpy

# assume on github test
try:
    from _common_helpers import has_display

    update = not has_display()
except ImportError:
    update = False


def test_use_batoms_startup():
    bpy.ops.batoms.delete()
    # no display, assume on github test
    if update:
        bpy.ops.batoms.use_batoms_startup()


def test_update():
    bpy.ops.batoms.delete()
    # no display, assume on github test
    if update:
        bpy.ops.batoms.update()


def test_pip_install_package():
    bpy.ops.batoms.delete()
    # no display, assume on github test
    if update:
        bpy.ops.batoms.pip_install_package()


def test_use_batoms_preference():
    bpy.ops.batoms.delete()
    # no display, assume on github test
    if update:
        bpy.ops.batoms.use_batoms_preference()



import bpy
from bpy.types import Panel
from bpy.props import (
    BoolProperty,
    IntProperty,
    FloatProperty,
    EnumProperty,
)

from batoms.utils.butils import get_selected_edges
from batoms.batoms import Batoms
from batoms.gui.utils import get_enum_attr


# The panel.
class Bond_PT_prepare(Panel):
    bl_label = "Edit bonds"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"DEFAULT_CLOSED"}
    bl_category = "Bond"
    bl_idname = "BATOMS_PT_Bond"
    bl_parent_id = "VIEW3D_PT_Batoms_bond"

    # @classmethod
    # def poll(cls, context):
    #     obj = context.object
    #     if obj:
    #         return obj.batoms.type == 'BOND' and obj.mode != 'EDIT'
    #     else:
    #         return False

    def draw(self, context):
        layout = self.layout
        bond = context.scene.Bbond

        layout.label(text="Bond style")
        col = layout.column()
        col.prop(bond, "style", expand=True)
        layout.prop(bond, "order")
        # layout.prop(bond, "width")
        layout.prop(bond, "show")


# ---------------------------------------------------
def get_active_bond():
    context = bpy.context
    if context.object and context.object.batoms.type == "BATOMS":
        indices = get_selected_edges(context.object)
        if len(indices) > 0:
            batoms = Batoms(label=context.object.batoms.label)
            slicebonds = batoms.bond[indices]
            return slicebonds
        else:
            return None
    return None


def set_bond_attr_by_batoms(key):
    """ """

    def setter(self, value):
        slicebonds = get_active_bond()
        if slicebonds is not None:
            # batoms = Batoms(label=slicebonds.label)
            # slicebonds = batoms.bond[slicebonds.indices]
            setattr(slicebonds, key, value)
            # bpy.ops.object.mode_set(mode="EDIT")
            bpy.context.view_layer.objects.active = slicebonds.obj

    return setter


def get_bond_attr(key):
    """ """

    def getter(self):
        bond = get_active_bond()
        if bond is not None:
            # batoms = Batoms(label=bond.label)
            # bpy.context.view_layer.objects.active = batoms.bond.obj
            value = getattr(bond, key)[0]
            # bpy.ops.object.mode_set(mode="EDIT")
            return value
        else:
            return self.bl_rna.properties[key].default

    return getter


class BondProperties(bpy.types.PropertyGroup):
    style: EnumProperty(
        name="style",
        description="bond style",
        items=(
            ("Unicolor cylinder", "Unicolor", "", 0),
            ("Bicolor cylinder", "Bicolor", "", 1),
            ("Dashed line", "Dashed", "", 2),
            ("Spring", "Spring", "", 3),
        ),
        default=1,
        get=get_enum_attr("style", get_active_bond),
        set=set_bond_attr_by_batoms("style"),
    )

    width: FloatProperty(
        name="bondwidth",
        default=0.1,
        min=0,
        soft_max=1,
        description="bondwidth",
        get=get_bond_attr("width"),
        set=set_bond_attr_by_batoms("width"),
    )

    order: IntProperty(
        name="Bond order",
        default=1,
        min=1,
        max=3,
        get=get_bond_attr("order"),
        set=set_bond_attr_by_batoms("order"),
    )
    show: BoolProperty(
        name="show",
        default=False,
        get=get_bond_attr("show"),
        set=set_bond_attr_by_batoms("show"),
    )



import bpy
import numpy as np

try:
    from _common_helpers import has_display, set_cycles_res

    use_cycles = not has_display()
except ImportError:
    use_cycles = False

extras = dict(engine="cycles") if use_cycles else {}


def test_settings(h2o_homo):
    """key search"""
    h2o = h2o_homo
    h2o.isosurface.settings["1"] = {"level": -0.001}
    h2o.isosurface.settings["2"] = {"level": 0.001, "color": [0, 0, 0.8, 0.5]}
    assert len(h2o.isosurface.settings) == 2
    assert h2o.isosurface.settings.find("1") is not None
    assert h2o.isosurface.settings.find(1) is not None
    h2o.isosurface.settings.remove("1")
    assert h2o.isosurface.settings.find("1") is None


# def test_slice():
#     bpy.ops.batoms.delete()
#     h2o = read("../tests/datas/h2o-homo.cube")
#     h2o.isosurface.settings["1"] = {"level": -0.001}
#     h2o.isosurface.settings["2"] = {"level": 0.001, "color": [0, 0, 0.8, 0.5]}
#     h2o.isosurface.draw()
#     h2o.lattice_plane.settings[(1, 0, 0)] = {"distance": 6, "slicing": True}
#     h2o.lattice_plane.draw()
#     if use_cycles:
#         set_cycles_res(h2o)
#     h2o.get_image([0, 0, 1], **extras)


def test_color_by(h2o_homo):
    from ase.io.cube import read_cube_data

    h2o = h2o_homo
    bpy.context.view_layer.objects.active = h2o.obj
    hartree, _atoms = read_cube_data("../tests/datas/h2o-hartree.cube")
    h2o.volumetric_data["hartree"] = -hartree
    bpy.ops.surface.isosurface_add(name="positive")
    h2o.isosurface.settings["positive"].level = 0.001
    h2o.isosurface.settings["positive"].color_by = "hartree"
    bpy.ops.surface.isosurface_draw()


def test_diff(h2o_homo):
    h2o = h2o_homo
    volume = h2o.volumetric_data["h2o_homo"]
    h2o.volumetric_data["h2o_homo"] = volume + 0.1
    assert np.allclose(h2o.volumetric_data["h2o_homo"], volume + 0.1)
    h2o.isosurface.settings["positive"] = {"level": 0.008, "color": [1, 1, 0, 0.8]}
    h2o.isosurface.settings["negative"] = {"level": -0.008, "color": [0, 0, 1, 0.8]}
    h2o.model_style = 1
    h2o.isosurface.draw()
    if use_cycles:
        set_cycles_res(h2o)
    else:
        h2o.render.resolution = [200, 200]
    h2o.get_image([0, 0, 1], output="h2o-homo-diff-top.png", **extras)
    h2o.get_image([1, 0, 0], output="h2o-homo-diff-side.png", **extras)


def test_isosurface_ops(h2o_homo):
    h2o = h2o_homo
    bpy.context.view_layer.objects.active = h2o.obj
    bpy.ops.surface.isosurface_draw()
    assert len(h2o.isosurface.settings) == 0
    bpy.ops.surface.isosurface_add(name="positive")
    assert len(h2o.isosurface.settings) == 1
    bpy.ops.surface.isosurface_draw()


def test_isosurface_uilist(h2o_homo):
    """isosurface panel"""
    h2o = h2o_homo
    bpy.context.view_layer.objects.active = h2o.obj
    h2o.obj.select_set(True)
    assert h2o.coll.Bisosurface.ui_list_index == 0
    bpy.ops.surface.isosurface_add(name="positive")
    assert h2o.coll.Bisosurface.ui_list_index == 0
    bpy.ops.surface.isosurface_add(name="negative")
    assert h2o.coll.Bisosurface.ui_list_index == 1



"""Definition of the Cavity class.

This module defines the Cavity object in the Batoms package.

"""

import bpy
import numpy as np
from batoms.base.object import ObjectGN
from batoms.plugins.base import PluginObject
from .setting import CavitySettings
from scipy import spatial
from batoms.utils.butils import get_node_by_name
from batoms.utils import string2Number
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)

default_attributes = [
    {"name": "species_index", "data_type": "INT"},
    {"name": "species", "data_type": "STRING"},
    {"name": "show", "data_type": "INT"},
    {"name": "scale", "data_type": "FLOAT"},
]

default_GroupInput = [
    ["species_index", "NodeSocketInt"],
    ["show", "NodeSocketBool"],
    ["scale", "NodeSocketFloat"],
]

default_cavity_datas = {
    "species_index": np.ones(0, dtype=int),
    "centers": np.zeros((1, 0, 3)),
    "show": np.zeros(0, dtype=int),
    "scale": np.ones(0, dtype=float),
}


class Cavity(ObjectGN, PluginObject):
    def __init__(
        self,
        label=None,
        cavity_datas=None,
        location=np.array([0, 0, 0]),
        batoms=None,
        resolution=None,
        minCave=None,
    ):
        """Cavity Class

        Args:
            label (_type_, optional): _description_. Defaults to None.
            location (_type_, optional): _description_. Defaults to np.array([0, 0, 0]).
            batoms (_type_, optional): _description_. Defaults to None.
        """
        #
        self.batoms = batoms
        self.label = label
        name = "cavity"
        ObjectGN.__init__(self, label, name)
        PluginObject.__init__(self, label, "cavity", batoms)
        if resolution is not None:
            self.resolution = resolution
        if minCave is not None:
            self.minCave = minCave
        flag = self.load()
        if not flag:
            if cavity_datas is None:
                cavity_datas = default_cavity_datas
            self.build_object(cavity_datas)
            self.settings = CavitySettings(self.label, batoms=batoms, parent=self)
        else:
            self.settings = CavitySettings(self.label, batoms=batoms, parent=self)
        self.settings.bpy_data.active = True

    def build_materials(self, name, color, node_inputs=None, material_style="default"):
        """ """
        from batoms.material import create_material

        if name in bpy.data.materials:
            mat = bpy.data.materials.get(name)
            bpy.data.materials.remove(mat, do_unlink=True)
        mat = create_material(
            name,
            color=color,
            node_inputs=node_inputs,
            material_style=material_style,
            backface_culling=False,
        )
        return mat

    def draw(self):
        """Draw cavity."""
        # delete old cavity
        cavity_datas = self.build_cavity()
        self.set_arrays(cavity_datas)

    def build_kdtree(self, positions):
        """ """
        from scipy import spatial

        # tstart = time()
        self.kdtree = spatial.KDTree(positions)
        self.kdtree_mesh = spatial.KDTree(self.meshgrids)
        # print('KDTree positions: %s' % (time() - tstart))

    def query_distance(self, points, parallel=1):
        # tstart = time()
        distance, indices = self.kdtree.query(points, workers=parallel)
        # print('KDTree query: %s' % (time() - tstart))
        # here we consider the size of atoms using one value for all.
        distance = distance - self.atomRadius
        return indices, distance

    def query_radius(self, meshgrids, points, radii, k=1):
        """
        Algorithm:
        Use KDTree to query the tree for neighbors within a radius r.
        """
        # tstart = time()
        kdtree_mesh = spatial.KDTree(meshgrids)
        indices = kdtree_mesh.query_ball_point(points, radii)
        # print('KDTree query: %s' % (time() - tstart))
        return indices[0]

    def base_meshgrids(self, grids):
        x, y, z = np.meshgrid(
            grids[0], grids[1], grids[2], indexing="ij"
        )  # , sparse=True)
        meshgrids = np.c_[x.ravel(), y.ravel(), z.ravel()]
        return meshgrids

    def build_grid(self, cell, resolution):
        """
        generate gridpoints
        """
        length = cell.length
        npoint = [int(x / resolution) for x in length]
        grids = []
        for i in range(3):
            grids.append(np.arange(npoint[i]) / npoint[i])
        meshgrids = self.base_meshgrids(grids)
        meshgrids = np.dot(meshgrids, cell.array)
        self.meshgrids = meshgrids
        self.shape = npoint

    def build_cavity(self):
        """Find cage base on radius range and boundary range
        Algorith:
            1) build a meshgrid
            2) calulated first neighbour distance by kdtree
            3) find the max distance (max radius sphere)
            4) remove meshgrid within this sphere,
            5) find next max radius, ..., untill radius < minCave
            6) remove all spheres contact with boundary
        Args:
            cell (_type_): _description_
            positions (_type_): _description_
            radius (_type_): _description_
            step (float, optional): _description_. Defaults to 1.0.
            boundary (list, optional): _description_. Defaults to [[0, 1], [0, 1], [0, 1]].

        Returns:
            _type_: _description_
        """
        from batoms.data import basic_colors

        arrays = self.batoms.arrays
        cell = self.batoms.cell
        self.build_grid(cell, self.resolution)
        self.build_kdtree(arrays["positions"])
        indices, distances = self.query_distance(self.meshgrids)
        spheres = self.find_cage_spheres(distances)
        spheres = self.check_sphere_boundary(spheres, cell)
        # spheres = self.refine_spheres(spheres)
        ns = len(spheres["centers"])
        color_names = list(basic_colors.keys())
        ic = 0
        # the default setting
        ncav = len(self.settings.bpy_setting)
        #
        indices = []
        for i in range(ns):
            has_r = False
            for cav in self.settings.bpy_setting:
                if spheres["radii"][i] > cav.min and spheres["radii"][i] < cav.max:
                    has_r = True
                    indices.append(i)
            # init setting if not exist (ncav == 0)
            if not has_r and ncav == 0:
                name = len(self.settings.bpy_setting)
                cav = {
                    "min": np.floor(spheres["radii"][i]),
                    "max": np.ceil(spheres["radii"][i]),
                    "color": basic_colors[color_names[ic]],
                }
                self.settings[name] = cav
                indices.append(i)
                ic += 1
                if ic == len(color_names):
                    ic = 0
        #
        cavities = {}
        for key, value in spheres.items():
            if len(spheres[key].shape) == 2:
                cavities[key] = spheres[key][indices, :]
            else:
                cavities[key] = spheres[key][indices]
        nc = len(cavities["centers"])
        species_index = np.zeros(nc, dtype=int)
        show = np.ones(nc, dtype=int)
        scale = np.ones(nc, dtype=int)
        ncollection = len(self.settings.bpy_setting)
        for i in range(ncollection):
            cav = self.settings.bpy_setting[i]
            indices = np.where(
                (cavities["radii"] > cav.min) & (cavities["radii"] < cav.max)
            )[0]
            species_index[indices] = string2Number(cav.name)
            scale[indices] = cavities["radii"][indices] * cav.scale
        # logger.debug(scale)
        cavities.update(
            {
                "centers": np.array([cavities["centers"]]),
                "species_index": species_index,
                "scale": scale,
                "show": show,
            }
        )
        return cavities

    def find_cage_spheres(self, distances):
        """
        Loop all sphere > min

        Args:
            distances (_type_): _description_

        Returns:
            _type_: _description_
        """
        meshgrids = self.meshgrids
        imax = np.argmax(distances)
        dmax = distances[imax]
        center = meshgrids[imax]
        # refine center
        center, dmax = self.refine_spheres(center)
        centers = []
        radii = []
        while dmax > self.minCave:
            centers.append(center)
            radii.append(dmax)
            n = len(distances)
            mask = np.ones(n, dtype=bool)
            indices1 = self.query_radius(meshgrids, [center], [dmax])
            mask[imax] = False
            mask[indices1] = False
            distances = distances[mask]
            meshgrids = meshgrids[mask]
            imax = np.argmax(distances)
            dmax = distances[imax]
            center = meshgrids[imax]
            center, dmax = self.refine_spheres(center)
            # print(center, dmax)
        spheres = {"centers": np.array(centers), "radii": np.array(radii)}
        # print('spheres: ', spheres)
        return spheres

    def check_sphere_boundary(self, spheres0, cell):
        """Remove sphere contact with boundary
        distance to cell < radius
        """
        from batoms.neighborlist import pointCellDistance

        centers = []
        radii = []
        if len(spheres0["centers"]) == 0:
            spheres = {"centers": np.array(centers), "radii": np.array(radii)}
            return spheres

        dis = pointCellDistance(spheres0["centers"], cell)
        # print('pointCellDistance: ', d)

        ns = len(spheres0["centers"])
        for i in range(ns):
            dmin = np.min(dis[:, :, i])
            # print(dmin, spheres0['radii'][i])
            if dmin > spheres0["radii"][i]:
                centers.append(spheres0["centers"][i])
                radii.append(spheres0["radii"][i])
        spheres = {"centers": np.array(centers), "radii": np.array(radii)}
        # print('spheres after check boundary: ', spheres)
        return spheres

    def refine_spheres(self, center):
        """Refine centers of spheres

        Find center off meshgrids.

        Args:
            spheres (_type_): _description_
        """
        # make dense meshgrid ground centers of spheres
        npoint = [5, 5, 5]
        grids = []
        for i in range(3):
            grid = (
                np.arange(-npoint[i], npoint[i] + 1) / npoint[i] * self.resolution / 2
            )
            grids.append(grid)
        meshgrids0 = self.base_meshgrids(grids)
        meshgrids = meshgrids0 + center
        indices, distances = self.query_distance(meshgrids)
        imax = np.argmax(distances)
        radius = distances[imax]
        return center, radius

    def build_object(self, cavity_datas, attributes={}):
        """Build the main Batoms object

        Args:
            label (str):
                Name of the object
            arrays (array):
                arrays of properties for each atoms
            location (list, optional):
                Location of the object. Defaults to [0, 0, 0].
        """
        if len(cavity_datas["centers"].shape) == 2:
            self._trajectory = {
                "centers": np.array([cavity_datas["centers"]]),
            }
            centers = cavity_datas["centers"]
        elif len(cavity_datas["centers"].shape) == 3:
            self._trajectory = {
                "centers": cavity_datas["centers"],
            }
            centers = cavity_datas["centers"][0]
        else:
            raise Exception("Shape of centers is wrong!")
        attributes.update(
            {
                "species_index": cavity_datas["species_index"],
                "show": cavity_datas["show"],
                "scale": cavity_datas["scale"],
            }
        )
        name = self.obj_name
        self.delete_obj(name)
        mesh = bpy.data.meshes.new(name)
        obj = bpy.data.objects.new(name, mesh)
        obj.data.from_pydata(centers, [], [])
        obj.batoms.type = "CAVITY"
        obj.batoms.label = self.batoms.label
        self.batoms.coll.objects.link(obj)
        # Add attributes
        for att in default_attributes:
            self.add_attribute(**att)
        # add cell object as its child
        obj.parent = self.batoms.obj
        self.set_attributes(attributes)
        self.init_geometry_node_modifier(default_GroupInput)
        self.build_geometry_node()

    def build_geometry_node(self):
        """Geometry node for instancing sphere on vertices!"""
        nodes = self.gn_node_group.nodes
        links = self.gn_node_group.links
        GroupInput = nodes[0]
        GroupOutput = nodes[1]
        # print(gn.name)
        JoinGeometry = get_node_by_name(
            nodes, "%s_JoinGeometry" % self.label, "GeometryNodeJoinGeometry"
        )
        SeparateGeometry = get_node_by_name(
            nodes, "%s_SeparateGeometry" % self.label, "GeometryNodeSeparateGeometry"
        )
        links.new(GroupInput.outputs["Geometry"], SeparateGeometry.inputs["Geometry"])
        links.new(GroupInput.outputs[2], SeparateGeometry.inputs["Selection"])
        links.new(SeparateGeometry.outputs[0], JoinGeometry.inputs["Geometry"])
        links.new(JoinGeometry.outputs["Geometry"], GroupOutput.inputs["Geometry"])

    def add_geometry_node(self, spname, instancer):
        """Add geometry node for each species

        Args:
            spname (str):
                Name of the species
            instancer (bpy.data.object):
                Object to be instanced
        """
        from batoms.utils.butils import compareNodeType

        nodes = self.gn_node_group.nodes
        links = self.gn_node_group.links
        GroupInput = nodes[0]
        JoinGeometry = get_node_by_name(
            nodes, "%s_JoinGeometry" % self.label, "GeometryNodeJoinGeometry"
        )
        CompareSpecies = get_node_by_name(
            nodes, "CompareFloats_%s_%s" % (self.label, spname), compareNodeType
        )
        CompareSpecies.operation = "EQUAL"
        # CompareSpecies.data_type = 'INT'
        CompareSpecies.inputs[1].default_value = string2Number(spname)
        InstanceOnPoint = get_node_by_name(
            nodes,
            "InstanceOnPoint_%s_%s" % (self.label, spname),
            "GeometryNodeInstanceOnPoints",
        )
        ObjectInfo = get_node_by_name(
            nodes, "ObjectInfo_%s_%s" % (self.label, spname), "GeometryNodeObjectInfo"
        )
        ObjectInfo.inputs["Object"].default_value = instancer
        BoolShow = get_node_by_name(
            nodes,
            "BooleanMath_%s_%s_1" % (self.label, spname),
            "FunctionNodeBooleanMath",
        )
        #
        links.new(GroupInput.outputs["Geometry"], InstanceOnPoint.inputs["Points"])
        links.new(GroupInput.outputs[1], CompareSpecies.inputs[0])
        links.new(GroupInput.outputs[2], BoolShow.inputs[0])
        links.new(GroupInput.outputs[3], InstanceOnPoint.inputs["Scale"])
        links.new(CompareSpecies.outputs[0], BoolShow.inputs[1])
        links.new(BoolShow.outputs["Boolean"], InstanceOnPoint.inputs["Selection"])
        links.new(ObjectInfo.outputs["Geometry"], InstanceOnPoint.inputs["Instance"])
        links.new(InstanceOnPoint.outputs["Instances"], JoinGeometry.inputs["Geometry"])

    def update_geometry_node_instancer(self):
        """ """
        for sp in self.settings.bpy_setting:
            self.settings.build_instancer(sp.as_dict())

    def set_arrays(self, arrays):
        """ """
        # if len(arrays['positions']) == 0:
        #     return
        attributes = self.attributes
        # same length
        dnvert = len(arrays["species_index"]) - len(attributes["species_index"])
        if dnvert > 0:
            self.add_vertices_bmesh(dnvert)
        elif dnvert < 0:
            self.delete_vertices_bmesh(range(-dnvert))
        self.set_trajectory(arrays)
        self.set_attributes({"species_index": arrays["species_index"]})
        self.set_attributes({"scale": arrays["scale"]})
        self.set_attributes({"show": arrays["show"]})
        self.update_mesh()
        self.update_geometry_node_instancer()

    def set_trajectory(self, trajectory=None, frame_start=0):
        if trajectory is None:
            trajectory = self._trajectory
        nframe = len(trajectory["centers"])
        if nframe == 0:
            return
        name = "%s_cavity" % (self.label)
        obj = self.obj
        self.set_shape_key(name, obj, trajectory["centers"], frame_start=frame_start)

    @property
    def objs(self):
        objs = {}
        name = "{}_{}".format(self.label, self.name)
        obj = bpy.data.objects.get(name)
        objs[name] = obj
        return objs

    @property
    def setting(self):
        from batoms.utils import deprecated

        """setting object."""
        deprecated(
            '"setting" will be deprecated in the furture, please use "settings".'
        )
        return self.settings

    def as_dict(self):
        """ """
        data = {}
        data["settings"] = self.settings.as_dict()
        data.update(self.settings.bpy_data.as_dict())
        return data

    @property
    def minCave(self):
        return self.batoms.coll.Bcavity.minCave

    @minCave.setter
    def minCave(self, minCave):
        self.batoms.coll.Bcavity.minCave = minCave

    @property
    def resolution(self):
        return self.batoms.coll.Bcavity.resolution

    @resolution.setter
    def resolution(self, resolution):
        self.batoms.coll.Bcavity.resolution = resolution

    @property
    def atomRadius(self):
        return self.batoms.coll.Bcavity.atomRadius

    @atomRadius.setter
    def atomRadius(self, atomRadius):
        self.batoms.coll.Bcavity.atomRadius = atomRadius



import os
from ase import io
from ase.io.cube import read_cube_data
from batoms import Batoms

# from time import time


def read(filename, label=None, **kwargs):
    """
    wrapper function for ase.io.read
    """
    base = os.path.basename(filename)
    base = os.path.splitext(base)
    if label is None:
        label = base[0]
        label = label.replace("-", "_")
        if label[:-1].isdigit():
            label = "b_" + label
    ext = base[1]
    if ext == ".cube":
        # tstart = time()
        volume, atoms = read_cube_data(filename, **kwargs)
        # print('Read cube: {0:1.2f}'.format(time() - tstart))
        batoms = Batoms(label, from_ase=atoms, volume={label: volume})
    elif "CHGCAR" in filename.upper():
        from pymatgen.io.vasp.outputs import VolumetricData

        poscar, data, data_aug = VolumetricData.parse_file("CHGCAR")
        # load structure and vlumetric data into Batoms
        batoms = Batoms("batoms", from_pymatgen=poscar.structure)
        batoms.volumetric_data["chgcar"] = data["total"]
    else:
        atoms = io.read(filename=filename, **kwargs)
        batoms = Batoms(label=label, from_ase=atoms)
    return batoms



"""VIEW3D_MT_edit_mesh_context_menu
"""

from bpy.types import Menu


def menu_func(self, context):
    self.layout.menu("VIEW3D_MT_edit_mesh_context_batoms", icon="MESH_UVSPHERE")


class VIEW3D_MT_edit_mesh_context_batoms_model_style(Menu):
    bl_idname = "VIEW3D_MT_edit_mesh_context_batoms_model_style"
    bl_label = "model style"
    # bl_icon = ""

    def draw(self, context):
        (
            is_vert_mode,
            is_edge_mode,
            is_face_mode,
        ) = context.tool_settings.mesh_select_mode

        layout = self.layout

        layout.operator_context = "INVOKE_REGION_WIN"

        op0 = layout.operator("batoms.apply_model_style_selected", text="Space-filling")
        op0.model_style = "0"
        op1 = layout.operator(
            "batoms.apply_model_style_selected", text="Ball-and-stick"
        )
        op1.model_style = "1"
        op2 = layout.operator("batoms.apply_model_style_selected", text="Polyhedral")
        op2.model_style = "2"
        op3 = layout.operator("batoms.apply_model_style_selected", text="Stick")
        op3.model_style = "3"


class VIEW3D_MT_edit_mesh_context_batoms(Menu):
    bl_idname = "VIEW3D_MT_edit_mesh_context_batoms"
    bl_label = "Batoms"
    bl_icon = "batoms_molecule"

    def draw(self, _context):
        layout = self.layout

        layout.operator_context = "INVOKE_REGION_WIN"
        layout.menu(
            "VIEW3D_MT_edit_mesh_context_batoms_model_style",
            text="Model Style",
            icon="LAYER_ACTIVE",
        )
        # layout.menu("VIEW3D_MT_edit_mesh_context_batoms_color_style",
        # text="Color Style", icon='LAYER_ACTIVE')
        # layout.menu("VIEW3D_MT_edit_mesh_context_batoms_radius_style",
        # text="Radius Style", icon='LAYER_ACTIVE')
        layout.operator("batoms.replace", text="Replace", icon="LAYER_ACTIVE")
        # layout.operator("batoms.select", text="Select", icon='LAYER_ACTIVE')
        layout.operator(
            "batoms.delete_selected_atoms", text="Delete", icon="FORCE_LENNARDJONES"
        )

        layout.separator()



import numpy as np
from batoms.utils import number2String
from batoms.base.object import childObjectGN
import logging

logger = logging.getLogger(__name__)


class SliceBonds(childObjectGN):
    """Batom Class"""

    def __init__(self, label, indices=0, bonds=None) -> None:
        childObjectGN.__init__(
            self, label, indices, obj_name="{}_bond".format(label), parent=bonds
        )

    def __repr__(self):
        # bpy.context.view_layer.objects.active = self.parent.obj
        # mode = self.parent.obj.mode
        # bpy.ops.object.mode_set(mode='OBJECT')
        s = "SliceBonds(species = %s, " % self.species["name"]
        s += "order = %s, " % self.order
        s += "style = %s)" % self.style
        # bpy.ops.object.mode_set(mode=mode)
        return s

    @property
    def species(self):
        # TODO: support slice
        sp1 = self.get_attribute("species_index0")
        sp2 = self.get_attribute("species_index1")
        if len(self.indices) == 1:
            name = np.array(["%s-%s" % (number2String(sp1), number2String(sp2))])
        else:
            n = len(sp1)
            name = np.zeros(n, dtype="U20")
            for i in range(n):
                name[i] = "%s-%s" % (number2String(sp1[i]), number2String(sp2[i]))
        order = self.get_attribute("bond_order")
        style = self.get_attribute("bond_style")
        # name = '%s-%s' % (sp1, sp2)
        sp = self.parent.settings[name[0]].as_dict()
        sp.update({"order": order[0], "style": style[0]})
        return sp

    @species.setter
    def species(self, species):
        self.set_species(species)

    def set_species(self, species):
        # self.attributes['species'].data[self.indices].value = species
        self.parent.replace([self.indices], species)

    @property
    def order(self):
        order = self.get_attribute("bond_order")
        return order

    @order.setter
    def order(self, value):
        self.set_attribute("bond_order", value)
        # find plane for high order
        a2, a3 = self.secondBond()
        self.set_attribute("atoms_index2", a2)
        self.set_attribute("atoms_index3", a3)
        # if order not exist, add one
        sp = self.species
        sp["order"] = value
        self.parent.settings.build_instancer(sp)
        self.parent.add_bond_pair_node(sp)

    @property
    def style(self):
        style = self.get_attribute("bond_style")
        return style

    @style.setter
    def style(self, value):
        self.set_attribute("bond_style", value)
        # if stule not exist, add one
        sp = self.species
        sp["style"] = value
        self.parent.settings.build_instancer(sp)
        self.parent.add_bond_pair_node(sp)

    def secondBond(self):
        """
        determine the plane of high order bond
        """
        arrays = self.parent.arrays
        ai = arrays["atoms_index0"][self.indices[0]]
        aj = arrays["atoms_index1"][self.indices[0]]
        indi = np.where(
            (arrays["atoms_index0"] == ai)  # noqa: W503
            | (arrays["atoms_index1"] == ai)  # noqa: W503
            | (arrays["atoms_index0"] == aj)  # noqa: W503
            | (arrays["atoms_index1"] == aj)  # noqa: W503
        )[0]
        if len(indi) == 1:
            second_bond = [0, 1]
        else:
            for i in indi:
                if arrays["atoms_index0"][i] == ai and arrays["atoms_index1"][i] == aj:
                    continue
                second_bond = [arrays["atoms_index0"][i], arrays["atoms_index1"][i]]
                break
        return second_bond

    def high_order_bond_plane(self):
        """
        determine the plane of high order bond
        """
        arrays = self.parent.arrays
        ai = arrays["atoms_index0"][self.indices]
        aj = arrays["atoms_index1"][self.indices]
        indi = np.where(arrays["atoms_index0"] == ai)[0]
        if len(indi) == 1:
            second_bond = np.array([0.0, 0.0, 1])
        else:
            for i in indi:
                if arrays["atoms_index1"][i] == aj:
                    continue
                second_bond = arrays["vectors"][i]
                break
        vec = np.cross(second_bond, arrays["vectors"][self.indices]) + np.array(
            [1e-8, 0, 0]
        )
        offset = np.cross(vec, arrays["vectors"][self.indices]) + np.array([1e-8, 0, 0])
        return offset



import bpy


def test_species(ch4):
    bpy.context.view_layer.objects.active = ch4.obj
    assert len(ch4.species) == 2
    bpy.ops.batoms.species_update()
    bpy.ops.batoms.species_add(species="O")
    assert len(ch4.species) == 3
    bpy.ops.batoms.species_remove(species="O")
    assert len(ch4.species) == 2
